# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the mesh creation functions."""

import re

import autograd.numpy as npAD
import numpy as np
import pytest
import splinepy
from autograd import jacobian

from beamme.core.mesh import Mesh
from beamme.core.node import NodeCosserat
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import Beam3rHerm2Line3, get_four_c_reissner_beam
from beamme.mesh_creation_functions.applications.beam_fibers_in_rectangle import (
    create_fibers_in_rectangle,
)
from beamme.mesh_creation_functions.applications.beam_stent import (
    create_beam_mesh_stent,
)
from beamme.mesh_creation_functions.applications.beam_wire import create_wire_fibers
from beamme.mesh_creation_functions.beam_arc import (
    create_beam_mesh_arc_segment_2d,
    create_beam_mesh_arc_segment_via_axis,
    create_beam_mesh_arc_segment_via_rotation,
)
from beamme.mesh_creation_functions.beam_generic import create_beam_mesh_generic
from beamme.mesh_creation_functions.beam_helix import create_beam_mesh_helix
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.mesh_creation_functions.beam_node_continuation import (
    create_beam_mesh_arc_at_node,
    create_beam_mesh_line_at_node,
)
from beamme.mesh_creation_functions.beam_parametric_curve import (
    create_beam_mesh_parametric_curve,
)
from beamme.mesh_creation_functions.beam_splinepy import (
    create_beam_mesh_from_splinepy,
    get_curve_function_and_jacobian_for_integration,
)
from beamme.utils.nodes import get_nodal_coordinates


def create_helix_function(
    radius, incline, *, transformation_factor=None, number_of_turns=None
):
    """Create and return a parametric function that represents a helix shape.
    The parameter coordinate can optionally be stretched to make the curve arc-
    length along the parameter coordinated non-constant and create a more
    complex curve for testing purposes.

    Args
    ----
    radius: float
        Radius of the helix
    incline: float
        Incline of the helix
    transformation_factor: float
        Factor to control the coordinate stretching (no direct physical interpretation)
    number_of_turns: float
        Number of turns the helix will have to get approximate boundaries for the transformation.
        This is only used for the transformation, not the actual geometry, as we return the
        function to create the geometry and not the geometry itself.
    """

    if transformation_factor is None and number_of_turns is None:

        def transformation(t):
            """Return identity transformation."""
            return 1.0

    elif transformation_factor is not None and number_of_turns is not None:

        def transformation(t):
            """Transform the parameter coordinate to make the function more
            complex."""
            return (
                npAD.exp(transformation_factor * t / (2.0 * np.pi * number_of_turns))
                * t
                / npAD.exp(transformation_factor)
            )

    else:
        raise ValueError(
            "You have to set none or both optional parameters: "
            "transformation_factor and number_of_turns"
        )

    def helix(t):
        """Parametric function to describe a helix."""
        return npAD.array(
            [
                radius * npAD.cos(transformation(t)),
                radius * npAD.sin(transformation(t)),
                transformation(t) * incline / (2 * np.pi),
            ]
        )

    return helix


def create_testing_bezier_curve():
    """Create a Bezier curve used for testing."""

    control_points = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 2.0, 1.0],
            [2.0, 2.0, 2.0],
            [3.0, 0.0, 0.0],
        ]
    )
    return splinepy.Bezier(degrees=[3], control_points=control_points)


def create_testing_nurbs_curve():
    """Create a NURBS curve used for testing."""

    return splinepy.NURBS(
        degrees=[2],
        knot_vectors=[[0, 0, 0, 1, 1, 1]],
        control_points=[[0, 0, 0], [1, 2, -1], [2, 0, 0]],
        weights=[[1.0], [1.0], [1.0]],
    )


def test_mesh_creation_functions_arc_segment_via_axis(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a circular segment via the axis method and compare it with the
    reference file."""

    # Create mesh
    mesh = Mesh()
    radius = 2.0
    beam_set = create_beam_mesh_arc_segment_via_axis(
        mesh,
        Beam3rHerm2Line3,
        get_default_test_beam_material(material_type="reissner"),
        [0, 0, 1],
        [0, radius, 0],
        [0, 0, 0],
        1.0,
        n_el=3,
    )
    mesh.add(beam_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_arc_segment_start_end_node(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Check that if start end nodes with non-matching positions or tangents
    are provided we get an error."""

    angle = 1.0
    radius = 2.0
    start_node_pos = [0, 0, 0]
    start_node_rot = Rotation()
    end_node_pos = radius * np.array([np.sin(angle), 1.0 - np.cos(angle), 0])
    end_node_rot = Rotation([0, 0, 1], angle)

    def create_beam(*, start_node=None, end_node=None):
        """This is the base function we use to generate the beam in this test
        case."""
        mesh = Mesh()
        mat = get_default_test_beam_material(material_type="reissner")
        if start_node is not None:
            mesh.add(start_node)
        if end_node is not None:
            mesh.add(end_node)
        create_beam_mesh_arc_segment_via_axis(
            mesh,
            Beam3rHerm2Line3,
            mat,
            [0, 0, 1],
            [0, radius, 0],
            [0, 0, 0],
            angle,
            n_el=3,
            start_node=start_node,
            end_node=end_node,
        )
        return mesh

    # This should work as expected, as all the values match.
    start_node = NodeCosserat(start_node_pos, start_node_rot)
    end_node = NodeCosserat(end_node_pos, end_node_rot)
    mesh = create_beam(start_node=start_node, end_node=end_node)
    assert_results_close(get_corresponding_reference_file_path(), mesh)

    # Create with start node where the position does not match.
    with pytest.raises(
        ValueError,
        match="start_node position does not match with function!",
    ):
        start_node = NodeCosserat([0, 1, 0], start_node_rot)
        create_beam(start_node=start_node)

    # Create with start node where the rotation does not match.
    with pytest.raises(
        ValueError,
        match="The tangent of the start node does not match with the given function!",
    ):
        start_node = NodeCosserat(start_node_pos, Rotation([1, 2, 3], np.pi / 3))
        create_beam(start_node=start_node)

    # Create with end node where the position does not match.
    with pytest.raises(
        ValueError,
        match="end_node position does not match with function!",
    ):
        end_node = NodeCosserat([0, 0, 0], end_node_rot)
        create_beam(end_node=end_node)

    # Create with end node where the rotation does not match.
    with pytest.raises(
        ValueError,
        match="The tangent of the end node does not match with the given function!",
    ):
        end_node = NodeCosserat(end_node_pos, Rotation([1, 2, 3], np.pi / 3))
        create_beam(end_node=end_node)


def test_mesh_creation_functions_arc_segment_via_rotation(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a circular segment via the rotation method and compare it with
    the reference file."""

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Create mesh.
    beam_set = create_beam_mesh_arc_segment_via_rotation(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [3, 6, 9.2],
        Rotation([4.5, 7, 10], np.pi / 5),
        10,
        np.pi / 2.3,
        n_el=5,
    )

    # Add boundary conditions.
    mesh.add(beam_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_arc_segment_2d(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a circular segments in 2D."""

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Create mesh.
    beam_set_1 = create_beam_mesh_arc_segment_2d(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [1.0, 2.0, 0.0],
        1.5,
        np.pi * 0.25,
        np.pi * (1.0 + 1.0 / 3.0),
        n_el=5,
    )
    beam_set_2 = create_beam_mesh_arc_segment_2d(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [1.0, 2.0, 0.0] - 2.0 * 0.5 * np.array([1, np.sqrt(3), 0]),
        0.5,
        np.pi / 3.0,
        -np.pi,
        n_el=3,
        start_node=mesh.nodes[-1],
    )

    # Add geometry sets
    mesh.add(beam_set_1, beam_set_2)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_node_positions_of_elements_option(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Creates a line, a circular segments in 2D and a helix by setting the
    node_positions_of_elements."""

    # Create a mesh
    mesh = Mesh()

    # Create and add material to mesh.
    material = get_default_test_beam_material(material_type="reissner")
    mesh.add(material)

    # Create a beam line with specified node_positions_of_elements.
    create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        material,
        [-1, -1, 0],
        [-1, -1, 3],
        node_positions_of_elements=[0, 0.1, 0.5, 0.9, 1.0],
    )

    # Create an arc segment similar to the equally spaced one,
    # but based on the provided node_positions_of_elements.
    create_beam_mesh_arc_segment_2d(
        mesh,
        Beam3rHerm2Line3,
        material,
        [1.0, 2.0, 0.0],
        1.5,
        np.pi * 0.25,
        np.pi * (1.0 + 1.0 / 3.0),
        node_positions_of_elements=[0.0, 0.2, 0.4, 0.6, 0.8, 1.0],
    )

    # Create a 2d segments with different element sizes.
    create_beam_mesh_arc_segment_2d(
        mesh,
        Beam3rHerm2Line3,
        material,
        [1.0, 2.0, 0.0],
        2.0,
        np.pi * 0.25,
        np.pi * (1.0 + 1.0 / 3.0),
        node_positions_of_elements=[0, 1.0 / 3.0, 1.0],
    )

    # Create a helix with different node positions.
    create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        material,
        [0.0, 0.0, 1.0],
        [4.0, 4.0, 0.0],
        [5.0, 5.0, 0.0],
        height_helix=10.0,
        turns=2.5 / np.pi,
        node_positions_of_elements=[0, 1.0 / 3.0, 1],
    )

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_stent(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the stent creation function."""

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Create mesh.
    create_beam_mesh_stent(
        mesh,
        Beam3rHerm2Line3,
        mat,
        0.11,
        0.02,
        5,
        8,
        fac_bottom=0.6,
        fac_neck=0.52,
        fac_radius=0.36,
        alpha=0.47 * np.pi,
        n_el=2,
    )

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_fibers_in_rectangle(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the create_fibers_in_rectangle function."""

    # Create mesh
    mesh = Mesh()

    # Create mesh.
    mat = get_default_test_beam_material(material_type="reissner")
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        45,
        0.45,
        0.35,
    )
    mesh.translate([0, 0, 1])
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        0,
        0.45,
        0.35,
    )
    mesh.translate([0, 0, 1])
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        90,
        0.45,
        0.35,
    )
    mesh.translate([0, 0, 1])
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        -90,
        0.45,
        0.35,
    )
    mesh.translate([0, 0, 1])
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        235,
        0.45,
        0.35,
    )
    mesh.translate([0, 0, 1])
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        1,
        4,
        30,
        0.45,
        5,
        fiber_element_length_min=0.2,
    )
    mesh.translate([0, 0, 1])
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        30,
        0.45,
        0.9,
    )
    mesh.translate([0, 0, 1])

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_fibers_in_rectangle_reference_point(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the create_fibers_in_rectangle function with using the
    reference_point option."""

    # Create mesh
    mesh = Mesh()

    # Create mesh.
    mat = get_default_test_beam_material(material_type="reissner")
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        45,
        0.45,
        0.35,
    )
    reference_point = 0.5 * np.array([4.0, 1.0]) + 0.1 * np.array(
        [-1.0, 1.0]
    ) / np.sqrt(2.0)
    create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        45,
        0.45,
        0.35,
        reference_point=reference_point,
    )

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_fibers_in_rectangle_return_set(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the set returned by the create_fibers_in_rectangle function."""

    # Create mesh
    mesh = Mesh()

    # Create mesh.
    mat = get_default_test_beam_material(material_type="reissner")
    beam_set = create_fibers_in_rectangle(
        mesh,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        4,
        1,
        45,
        0.45,
        0.35,
    )
    mesh.add(beam_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_wire(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the create_wire_fibers function."""

    # Create mesh
    mesh = Mesh()

    # Create two wires with different parameters.
    mat = get_default_test_beam_material(material_type="reissner")
    mesh_1 = Mesh()
    set_1 = create_wire_fibers(
        mesh_1,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        3.0,
        layers=2,
        n_el=2,
        radius=0.05,
    )
    mesh_2 = Mesh()
    set_2 = create_wire_fibers(
        mesh_2,
        get_four_c_reissner_beam(n_nodes=2, is_hermite_centerline=False),
        mat,
        3.0,
        layers=2,
        n_el=2,
        radius=0.1,
    )
    mesh_2.translate([0.0, 1.5, 0.0])
    mesh.add(mesh_1, mesh_2, set_1, set_2)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


@pytest.mark.parametrize(
    ("name", "curve_creation_function", "ref_length"),
    [
        ("bezier", create_testing_bezier_curve, 5.064502358928783),
        ("nurbs", create_testing_nurbs_curve, 3.140204411551537),
    ],
)
def test_mesh_creation_functions_splinepy(
    name,
    curve_creation_function,
    ref_length,
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the create_beam_mesh_from_splinepy function with different splinepy
    curves."""

    curve = curve_creation_function()
    mat = get_default_test_beam_material(material_type="reissner")
    mesh = Mesh()
    _, length = create_beam_mesh_from_splinepy(
        mesh, Beam3rHerm2Line3, mat, curve, n_el=3, output_length=True
    )
    assert_results_close(ref_length, length)

    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier=name), mesh
    )


def test_mesh_creation_functions_splinepy_unit(assert_results_close):
    """Unittest the function and jacobian creation in the
    create_beam_mesh_from_splinepy function."""

    curve = create_testing_nurbs_curve()
    r, dr, _, _ = get_curve_function_and_jacobian_for_integration(curve, tol=10)

    t_values = [5.0 / 7.0, -0.3, 1.2]
    results_r = [
        [1.4285714285714286, 0.8163265306122449, -0.4081632653061225],
        [-0.6, -1.2, 0.6],
        [2.4, -0.8, 0.4],
    ]
    results_dr = [
        [2.0, -1.7142857142857144, 0.8571428571428572],
        [2.0, 4.0, -2.0],
        [2.0, -4.0, 2.0],
    ]

    for t, result_r, result_dr in zip(t_values, results_r, results_dr):
        assert_results_close(r(t), result_r)
        assert_results_close(dr(t), result_dr)


def test_mesh_creation_functions_node_continuation(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test that the node continuation function work as expected."""

    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")
    mesh.add(mat)

    start_node = NodeCosserat([1, 2, 3], Rotation())
    mesh.add(start_node)
    beam_set = create_beam_mesh_line_at_node(
        mesh, Beam3rHerm2Line3, mat, start_node, 1.2, n_el=1
    )
    beam_set = create_beam_mesh_arc_at_node(
        mesh,
        Beam3rHerm2Line3,
        mat,
        beam_set["end"],
        [0, 1, 0],
        1.0,
        np.pi * 0.5,
        n_el=1,
    )
    beam_set = create_beam_mesh_arc_at_node(
        mesh,
        Beam3rHerm2Line3,
        mat,
        beam_set["end"],
        [0, 1, 0],
        1.0,
        -np.pi * 0.5,
        n_el=2,
    )
    beam_set = create_beam_mesh_arc_at_node(
        mesh,
        Beam3rHerm2Line3,
        mat,
        beam_set["end"],
        [0, 0, 1],
        1.0,
        -np.pi * 3.0 / 4.0,
        n_el=1,
    )
    create_beam_mesh_line_at_node(
        mesh, Beam3rHerm2Line3, mat, beam_set["end"], 2.3, n_el=3
    )

    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_node_continuation_accumulated(
    get_default_test_beam_material, assert_results_close
):
    """Test that the arc node continuation function can be applied multiple
    times in a row.

    This function can lead to accumulated errors in the rotations if not
    implemented carefully.
    """

    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")
    mesh.add(mat)

    n_segments = 100

    rotation_ref = Rotation([1, 0, 0], 0.5 * np.pi) * Rotation([0, 0, 1], 0.5 * np.pi)
    start_node = NodeCosserat([1, 2, 3], rotation_ref)
    mesh.add(start_node)
    beam_set = {"end": start_node}
    angle = np.pi
    angle_increment = angle / n_segments
    axis = [1, 0, 0]
    for i in range(n_segments):
        beam_set = create_beam_mesh_arc_at_node(
            mesh,
            Beam3rHerm2Line3,
            mat,
            beam_set["end"],
            axis,
            1.0,
            angle_increment,
            n_el=2,
        )

    rotation_actual = beam_set["end"].get_points()[0].rotation

    # Calculate the solution and get the "analytical" solution
    rotation_expected = Rotation(axis, angle) * rotation_ref
    quaternion_expected = np.array([-0.5, 0.5, -0.5, -0.5])
    assert rotation_actual == rotation_expected
    assert_results_close(rotation_actual.q, quaternion_expected)


def test_mesh_creation_functions_element_length_option(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test that the element length can be specified in the beam creation
    functions."""

    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")

    l_el = 1.5

    mesh_line = Mesh()
    create_beam_mesh_line(
        mesh_line,
        Beam3rHerm2Line3,
        mat,
        [1.0, 2.0, 0.0],
        [3.0, 4.0, 6.0],
        l_el=l_el,
    )

    mesh_line_long = Mesh()
    create_beam_mesh_line(
        mesh_line_long,
        Beam3rHerm2Line3,
        mat,
        [1.0, 2.0, 2.0],
        [3.0, 4.0, 8.0],
        l_el=100,
    )

    mesh_arc = Mesh()
    create_beam_mesh_arc_segment_via_rotation(
        mesh_arc,
        Beam3rHerm2Line3,
        mat,
        [1.0, 2.0, 3.0],
        Rotation([1, 3, 4], np.pi / 3.0),
        2.0,
        np.pi * 2.0 / 3.0,
        l_el=l_el,
    )

    # Get the helix curve function
    R = 2.0
    tz = 4.0  # incline
    n = 0.5  # number of turns
    helix = create_helix_function(R, tz, transformation_factor=2.0, number_of_turns=n)

    mesh_curve = Mesh()
    create_beam_mesh_parametric_curve(
        mesh_curve,
        Beam3rHerm2Line3,
        mat,
        helix,
        [0.0, 2.0 * np.pi * n],
        l_el=l_el,
    )

    # Check the output
    mesh.add(mesh_line, mesh_line_long, mesh_arc, mesh_curve)
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_argument_checks(
    get_default_test_beam_material,
):
    """Test that wrong input values leads to failure."""

    dummy_arg = "dummy"

    # Check error messages for input parameters
    with pytest.raises(
        ValueError,
        match='The arguments "n_el", "l_el" and "node_positions_of_elements" are mutually exclusive',
    ):
        mesh = Mesh()
        # This should raise an error since we dont allow `n_el` and `l_el`
        # to be set at the same time.
        create_beam_mesh_line(
            mesh,
            Beam3rHerm2Line3,
            get_default_test_beam_material(material_type="reissner"),
            [1.0, 2.0, 0.0],
            [3.0, 4.0, 6.0],
            n_el=1,
            l_el=1.5,
        )
    with pytest.raises(
        ValueError,
        match='The arguments "n_el", "l_el" and "node_positions_of_elements" are mutually exclusive',
    ):
        mesh = Mesh()
        # This should raise an error because node_positions_of_elements can not be used with l_el.

        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            l_el=1,
            node_positions_of_elements=[0.0, 0.5, 1.0],
        )

    with pytest.raises(
        ValueError,
        match='The arguments "n_el", "l_el" and "node_positions_of_elements" are mutually exclusive',
    ):
        mesh = Mesh()
        # This should raise an error because node_positions_of_elements can not be used with n_el.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            n_el=1,
            node_positions_of_elements=[0.0, 0.5, 1.0],
        )

    with pytest.raises(
        ValueError, match='The parameter "l_el" requires "interval_length" to be set.'
    ):
        mesh = Mesh()
        # This should raise an error because we set `l_el` but don't provide
        # `interval_length`.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=[0, 1],
            l_el=2.0,
        )

    with pytest.raises(
        ValueError,
        match="First entry of node_positions_of_elements must be 0, got -1.0",
    ):
        mesh = Mesh()
        # This should raise an error because the interval [0,1] is violated.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            node_positions_of_elements=[-1.0, 0.0, 1.0],
        )

    with pytest.raises(
        ValueError, match="Last entry of node_positions_of_elements must be 1, got 2.0"
    ):
        mesh = Mesh()
        # This should raise an error because the interval [0,1] is violated.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            node_positions_of_elements=[0.0, 1.0, 2.0],
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "The given node_positions_of_elements must be in ascending order. Got [0.0, 0.2, 0.1, 1.0]"
        ),
    ):
        mesh = Mesh()
        # This should raise an error because the interval is not ordered.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            node_positions_of_elements=[0.0, 0.2, 0.1, 1.0],
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            'The arguments "close_beam" and "end_node" are mutually exclusive'
        ),
    ):
        mesh = Mesh()
        # This should raise an error because the interval is not ordered.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            n_el=1,
            close_beam=True,
            end_node=dummy_arg,
        )


@pytest.mark.parametrize(
    "basic_creation_function",
    ["line", "arc"],
)
def test_mesh_creation_functions_arc_length(
    basic_creation_function, get_default_test_beam_material, assert_results_close
):
    """Test that the arc length can be stored in the nodes when creating a
    filament."""

    node_positions_of_elements = [0, 0.25, 1]
    mat = get_default_test_beam_material(material_type="reissner")
    offset = 3.0
    if basic_creation_function == "line":
        length = 2.5
        start_pos = [0, 0, 0]
        end_pos = [length, 0, 0]
        start_rot = Rotation()
        end_rot = Rotation()

        def create_beam(mesh, **kwargs):
            """Wrapper for the common arguments in the call to create the
            line."""

            create_beam_mesh_line(
                mesh,
                Beam3rHerm2Line3,
                mat,
                start_pos,
                end_pos,
                node_positions_of_elements=node_positions_of_elements,
                **kwargs,
            )

    elif basic_creation_function == "arc":
        length = np.pi * 0.5
        start_pos = [0, 1, 0]
        end_pos = [-1, 0, 0]
        start_rot = Rotation([0, 0, 1], np.pi)
        end_rot = Rotation([0, 0, 1], 1.5 * np.pi)

        def create_beam(mesh, **kwargs):
            """Wrapper for the common arguments in the call to create the
            arc."""

            create_beam_mesh_arc_segment_2d(
                mesh,
                Beam3rHerm2Line3,
                mat,
                [0, 0, 0],
                1.0,
                np.pi * 0.5,
                np.pi,
                node_positions_of_elements=node_positions_of_elements,
                **kwargs,
            )

    arc_length_ref = np.array([0, 0.125, 0.25, 0.625, 1.0]) * length
    arc_length_offset_ref = arc_length_ref + offset

    def get_start_and_end_node(*, arc_length_start=None, arc_length_end=None):
        """Return the explicitly created start and end node."""
        start_node = NodeCosserat(start_pos, start_rot, arc_length=arc_length_start)
        end_node = NodeCosserat(end_pos, end_rot, arc_length=arc_length_end)
        return (start_node, end_node)

    def check_arc_length(mesh, arc_length_ref):
        """Compare the arc lengths of the nodes in mesh with reference
        values."""
        arc_length_from_mesh = np.array([node.arc_length for node in mesh.nodes])
        assert_results_close(
            {"arc_length": arc_length_from_mesh}, {"arc_length": arc_length_ref}
        )

    # Standard arc length calculation
    mesh = Mesh()
    create_beam(mesh, set_nodal_arc_length=True)
    check_arc_length(mesh, arc_length_ref)

    # Standard arc length calculation with offset
    mesh = Mesh()
    create_beam(mesh, set_nodal_arc_length=True, nodal_arc_length_offset=offset)
    check_arc_length(mesh, arc_length_offset_ref)

    # Arc length calculation based on start node
    mesh = Mesh()
    start_node, _ = get_start_and_end_node(arc_length_start=offset)
    mesh.add(start_node)
    create_beam(mesh, set_nodal_arc_length=True, start_node=start_node)
    check_arc_length(mesh, arc_length_offset_ref)

    # Arc length calculation based on end node
    mesh = Mesh()
    _, end_node = get_start_and_end_node(arc_length_end=offset + length)
    mesh.add(end_node)
    create_beam(mesh, set_nodal_arc_length=True, end_node=end_node)
    # The nodes are in different order here, so we need to reorder the reference result
    check_arc_length(
        mesh, [arc_length_offset_ref[-1]] + arc_length_offset_ref.tolist()[:-1]
    )

    # Arc length calculation based on start and end node (if the values don't
    # match an error will be raised)
    mesh = Mesh()
    start_node, end_node = get_start_and_end_node(
        arc_length_start=offset, arc_length_end=offset + length
    )
    mesh.add(start_node, end_node)
    create_beam(
        mesh, set_nodal_arc_length=True, start_node=start_node, end_node=end_node
    )
    # The nodes are in different order here, so we need to reorder the reference result
    check_arc_length(
        mesh,
        [arc_length_offset_ref[0]]
        + [arc_length_offset_ref[-1]]
        + arc_length_offset_ref.tolist()[1:-1],
    )

    # Arc length calculation based on all possible parameters (if the values don't
    # match an error will be raised)
    mesh = Mesh()
    start_node, end_node = get_start_and_end_node(
        arc_length_start=offset, arc_length_end=offset + length
    )
    mesh.add(start_node, end_node)
    create_beam(
        mesh,
        set_nodal_arc_length=True,
        start_node=start_node,
        end_node=end_node,
        nodal_arc_length_offset=offset,
    )
    # The nodes are in different order here, so we need to reorder the reference result
    check_arc_length(
        mesh,
        [arc_length_offset_ref[0]]
        + [arc_length_offset_ref[-1]]
        + arc_length_offset_ref.tolist()[1:-1],
    )


def test_mesh_creation_functions_arc_length_argument_checks(
    get_default_test_beam_material,
):
    """Test that wrong input values leads to failure."""

    dummy_arg = "dummy"

    # Check error messages for input parameters
    with pytest.raises(
        ValueError,
        match='The arguments "n_el", "l_el" and "node_positions_of_elements" are mutually exclusive',
    ):
        mesh = Mesh()
        # This should raise an error since we dont allow `n_el` and `l_el`
        # to be set at the same time.
        create_beam_mesh_line(
            mesh,
            Beam3rHerm2Line3,
            get_default_test_beam_material(material_type="reissner"),
            [1.0, 2.0, 0.0],
            [3.0, 4.0, 6.0],
            n_el=1,
            l_el=1.5,
        )
    with pytest.raises(
        ValueError,
        match='The arguments "n_el", "l_el" and "node_positions_of_elements" are mutually exclusive',
    ):
        mesh = Mesh()
        # This should raise an error because node_positions_of_elements can not be used with l_el.

        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            l_el=1,
            node_positions_of_elements=[0.0, 0.5, 1.0],
        )

    with pytest.raises(
        ValueError,
        match='The arguments "n_el", "l_el" and "node_positions_of_elements" are mutually exclusive',
    ):
        mesh = Mesh()
        # This should raise an error because node_positions_of_elements can not be used with n_el.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            n_el=1,
            node_positions_of_elements=[0.0, 0.5, 1.0],
        )

    with pytest.raises(
        ValueError, match='The parameter "l_el" requires "interval_length" to be set.'
    ):
        mesh = Mesh()
        # This should raise an error because we set `l_el` but don't provide
        # `interval_length`.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=[0, 1],
            l_el=2.0,
        )

    with pytest.raises(
        ValueError,
        match="First entry of node_positions_of_elements must be 0, got -1.0",
    ):
        mesh = Mesh()
        # This should raise an error because the interval [0,1] is violated.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            node_positions_of_elements=[-1.0, 0.0, 1.0],
        )

    with pytest.raises(
        ValueError, match="Last entry of node_positions_of_elements must be 1, got 2.0"
    ):
        mesh = Mesh()
        # This should raise an error because the interval [0,1] is violated.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            node_positions_of_elements=[0.0, 1.0, 2.0],
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "The given node_positions_of_elements must be in ascending order. Got [0.0, 0.2, 0.1, 1.0]"
        ),
    ):
        mesh = Mesh()
        # This should raise an error because the interval is not ordered.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            node_positions_of_elements=[0.0, 0.2, 0.1, 1.0],
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            'The arguments "close_beam" and "end_node" are mutually exclusive'
        ),
    ):
        mesh = Mesh()
        # This should raise an error because the interval is not ordered.
        create_beam_mesh_generic(
            mesh,
            beam_class=dummy_arg,
            material=dummy_arg,
            function_generator=dummy_arg,
            interval=dummy_arg,
            n_el=1,
            close_beam=True,
            end_node=dummy_arg,
        )


def test_mesh_creation_functions_curve_3d_helix(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a helix from a parametric curve where the parameter is
    transformed so the arc length along the beam is not proportional to the
    parameter."""

    # Create mesh
    mesh = Mesh()

    # Add material and functions.
    mat = get_default_test_beam_material(material_type="reissner")

    # Get the helix curve function
    R = 2.0
    tz = 4.0  # incline
    n = 1  # number of turns
    n_el = 5
    helix = create_helix_function(R, tz, transformation_factor=2.0, number_of_turns=n)

    helix_set = create_beam_mesh_parametric_curve(
        mesh, Beam3rHerm2Line3, mat, helix, [0.0, 2.0 * np.pi * n], n_el=n_el
    )
    mesh.add(helix_set)

    # Compare the coordinates with the ones from Mathematica.
    coordinates_mathematica = np.loadtxt(
        get_corresponding_reference_file_path(
            additional_identifier="mathematica", extension="csv"
        ),
        delimiter=",",
    )
    assert_results_close(coordinates_mathematica, get_nodal_coordinates(mesh.nodes))

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_curve_3d_helix_length(
    get_default_test_beam_material, assert_results_close
):
    """Create a helix from a parametric curve where and check that the correct
    length is returned."""

    mesh_1 = Mesh()
    mesh_2 = Mesh()
    mat = get_default_test_beam_material(material_type="reissner")

    # Get the helix curve function
    R = 2.0
    tz = 4.0  # incline
    n = 1  # number of turns
    n_el = 3
    helix = create_helix_function(R, tz, transformation_factor=2.0, number_of_turns=n)

    args = [Beam3rHerm2Line3, mat, helix, [0.0, 2.0 * np.pi * n]]
    kwargs = {"n_el": n_el}

    helix_set_1 = create_beam_mesh_parametric_curve(mesh_1, *args, **kwargs)
    mesh_1.add(helix_set_1)

    helix_set_2, length = create_beam_mesh_parametric_curve(
        mesh_2, *args, output_length=True, **kwargs
    )
    mesh_2.add(helix_set_2)

    # Check the computed length
    assert_results_close(length, 13.18763323790246)

    # Check that both meshes are equal
    assert_results_close(mesh_1, mesh_2)


def test_mesh_creation_functions_curve_2d_sin(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a sin from a parametric curve."""

    # Create mesh
    mesh = Mesh()

    # Add material and functions.
    mat = get_default_test_beam_material(material_type="reissner")

    # Set parameters for the sin.
    n_el = 8

    def sin(t):
        """Parametric curve as a sin."""
        return npAD.array([t, npAD.sin(t)])

    sin_set = create_beam_mesh_parametric_curve(
        mesh, Beam3rHerm2Line3, mat, sin, [0.0, 2.0 * np.pi], n_el=n_el
    )
    mesh.add(sin_set)

    # Compare the coordinates with the ones from Mathematica.
    coordinates_mathematica = np.loadtxt(
        get_corresponding_reference_file_path(
            additional_identifier="mathematica", extension="csv"
        ),
        delimiter=",",
    )
    assert_results_close(coordinates_mathematica, get_nodal_coordinates(mesh.nodes))

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_curve_3d_curve_rotation(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a line from a parametric curve and prescribe the rotation."""

    # Create mesh
    mesh = Mesh()

    # Add material and functions.
    mat = get_default_test_beam_material(material_type="reissner")

    # Set parameters for the line.
    L = 1.1
    n_el = 4

    def curve(t):
        """Parametric representation of the centerline curve."""
        return npAD.array([L * t, t * t * L * L, 0.0])

    def rotation(t):
        """Function that defines the triad along the centerline curve."""
        rp2 = jacobian(curve)(t)
        rp = [rp2[0], rp2[1], 0]
        R1 = Rotation([1, 0, 0], t * 2 * np.pi)
        R2 = Rotation.from_basis(rp, [0, 0, 1])
        return R2 * R1

    sin_set = create_beam_mesh_parametric_curve(
        mesh,
        Beam3rHerm2Line3,
        mat,
        curve,
        [0.0, 1.0],
        n_el=n_el,
        function_rotation=rotation,
    )
    mesh.add(sin_set)

    # extend test case with different meshing strategy
    create_beam_mesh_parametric_curve(
        mesh,
        Beam3rHerm2Line3,
        mat,
        curve,
        [1.0, 2.5],
        node_positions_of_elements=[0, 1.0 / 1.3, 1.0],
        function_rotation=rotation,
    )

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_curve_3d_line(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a line from a parametric curve.

    Once the interval is in ascending order, once in descending. This
    tests checks that the elements are created with the correct tangent
    vectors.
    """

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    def line(t):
        """Create a line with a parametric curve (and a transformed
        parameter)."""
        factor = 2
        t_trans = npAD.exp(factor * t / (2.0 * np.pi)) * t / npAD.exp(factor)
        return npAD.array([t_trans, 0, 0])

    # Create mesh.
    set_1 = create_beam_mesh_parametric_curve(
        mesh, Beam3rHerm2Line3, mat, line, [0.0, 5.0], n_el=3
    )
    mesh.translate([0, 1, 0])
    set_2 = create_beam_mesh_parametric_curve(
        mesh, Beam3rHerm2Line3, mat, line, [5.0, 0.0], n_el=3
    )
    mesh.add(set_1, set_2)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_helix_no_rotation(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a helix and compare it with the reference file."""

    ## Helix angle and height helix combination
    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Add simple line to verify that the helix creation does not alter additional meshes
    create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, mat, [0.0, 0.0, 0.0], [-1.0, -1.0, -1.0]
    )

    # Create helix.
    helix_set = create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0.0, 0.0, 1.0],
        [0.0, 0.0, 0.0],
        [2.0, 0.0, 0.0],
        helix_angle=np.pi / 4,
        height_helix=10.0,
        l_el=5.0,
    )
    mesh.add(helix_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)

    ## Helix angle and turns
    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Add simple line to verify that the helix creation does not alter additional meshes
    create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, mat, [0.0, 0.0, 0.0], [-1.0, -1.0, -1.0]
    )

    # Create helix.
    helix_set = create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0.0, 0.0, 1.0],
        [0.0, 0.0, 0.0],
        [2.0, 0.0, 0.0],
        helix_angle=np.pi / 4,
        turns=2.5 / np.pi,
        l_el=5.0,
    )
    mesh.add(helix_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Add simple line to verify that the helix creation does not alter additional meshes
    create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, mat, [0.0, 0.0, 0.0], [-1.0, -1.0, -1.0]
    )

    # Create helix.
    helix_set = create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0.0, 0.0, 1.0],
        [0.0, 0.0, 0.0],
        [2.0, 0.0, 0.0],
        height_helix=10.0,
        turns=2.5 / np.pi,
        l_el=5.0,
    )
    mesh.add(helix_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_helix_rotation_offset(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a helix and compare it with the reference file."""

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Add simple line to verify that the helix creation does not alter additional meshes
    create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, mat, [0.0, 0.0, 0.0], [-1.0, -1.0, -1.0]
    )

    # Create helix.
    helix_set = create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [1.0, 1.0, 1.0],
        [-1.0, -1.0, -1.0],
        [3.0, 0.0, 0.0],
        helix_angle=np.pi / 6,
        height_helix=10.0,
        l_el=5.0,
    )
    mesh.add(helix_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_helix_radius_zero(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a helix and compare it with the reference file."""

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Add simple line to verify that the helix creation does not alter additional meshes
    create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, mat, [0.0, 0.0, 0.0], [-1.0, -1.0, -1.0]
    )

    # Create helix.
    helix_set = create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [1.0, 1.0, 1.0],
        [-1.0, -1.0, -1.0],
        [1.0, 1.0, 1.0],
        helix_angle=np.pi / 6,
        height_helix=80.0,
        n_el=4,
        warning_straight_line=False,
    )
    mesh.add(helix_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_mesh_creation_functions_helix_helix_angle_right_angle(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a helix and compare it with the reference file."""

    # Create mesh
    mesh = Mesh()

    # Add material and function.
    mat = get_default_test_beam_material(material_type="reissner")

    # Add simple line to verify that the helix creation does not alter additional meshes
    create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, mat, [0.0, 0.0, 0.0], [-1.0, -1.0, -1.0]
    )

    # Create helix.
    helix_set = create_beam_mesh_helix(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [1.0, 1.0, 1.0],
        [-1.0, -1.0, -1.0],
        [2.0, 2.0, 1.0],
        helix_angle=np.pi / 2,
        height_helix=10.0,
        l_el=5.0,
        warning_straight_line=False,
    )
    mesh.add(helix_set)

    # Check the output.
    assert_results_close(get_corresponding_reference_file_path(), mesh)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module defines a class that represents a rotation in 3D."""

import numpy as _np
import quaternion as _quaternion
from numpy.typing import NDArray as _NDArray

from beamme.core.conf import bme as _bme


def skew_matrix(vector):
    """Return the skew matrix for the vector."""
    skew = _np.zeros([3, 3])
    skew[0, 1] = -vector[2]
    skew[0, 2] = vector[1]
    skew[1, 0] = vector[2]
    skew[1, 2] = -vector[0]
    skew[2, 0] = -vector[1]
    skew[2, 1] = vector[0]
    return skew


class Rotation:
    """A class that represents a rotation of a coordinate system.

    Internally the rotations are stored as quaternions.
    """

    def __init__(self, *args):
        """Initialize the rotation object.

        Args
        ----
        *args:
            - Rotation()
                Create a identity rotation.
            - Rotation(axis, phi)
                Create a rotation around the vector axis with the angle phi.
        """

        self.q = _np.zeros(4)

        if len(args) == 0:
            # Identity element.
            self.q[0] = 1
        elif len(args) == 2:
            # Set from rotation axis and rotation angle.
            axis = _np.asarray(args[0])
            phi = args[1]
            norm = _np.linalg.norm(axis)
            if norm < _bme.eps_quaternion:
                raise ValueError("The rotation axis can not be a zero vector!")
            self.q[0] = _np.cos(0.5 * phi)
            self.q[1:] = _np.sin(0.5 * phi) * axis / norm
        else:
            raise ValueError(f"The given arguments {args} are invalid!")

    @classmethod
    def from_quaternion(cls, q, *, normalized=False):
        """Create the object from a quaternion float array (4x1)

        Args
        ----
        q: Quaternion, q0, qx,qy,qz
        normalized: Flag if the input quaternion is normalized. If so, no
            normalization is performed which can potentially improve performance.
            Skipping the normalization should only be done in very special cases
            where we can be sure that the input quaternion is normalized to avoid
            error accumulation.
        """
        if isinstance(q, _quaternion.quaternion):
            q = _quaternion.as_float_array(q)

        rotation = object.__new__(cls)
        if normalized:
            rotation.q = _np.array(q)
        else:
            rotation.q = _np.asarray(q) / _np.linalg.norm(q)
        if (not rotation.q.ndim == 1) or (not len(rotation.q) == 4):
            raise ValueError("Got quaternion array with unexpected dimensions")
        return rotation

    @classmethod
    def from_rotation_matrix(cls, R):
        """Create the object from a rotation matrix.

        The code is based on Spurriers algorithm:
            R. A. Spurrier (1978): “Comment on “Singularity-free extraction of a quaternion from a
            direction-cosine matrix”
        """

        R = _np.asarray(R)
        q = _np.zeros(4)
        trace = _np.trace(R)
        values = [R[i, i] for i in range(3)]
        values.append(trace)
        arg_max = _np.argmax(values)
        if arg_max == 3:
            q[0] = _np.sqrt(trace + 1) * 0.5
            q[1] = (R[2, 1] - R[1, 2]) / (4 * q[0])
            q[2] = (R[0, 2] - R[2, 0]) / (4 * q[0])
            q[3] = (R[1, 0] - R[0, 1]) / (4 * q[0])
        else:
            i_index = arg_max
            j_index = (i_index + 1) % 3
            k_index = (i_index + 2) % 3
            q_i = _np.sqrt(R[i_index, i_index] * 0.5 + (1 - trace) * 0.25)
            q[0] = (R[k_index, j_index] - R[j_index, k_index]) / (4 * q_i)
            q[i_index + 1] = q_i
            q[j_index + 1] = (R[j_index, i_index] + R[i_index, j_index]) / (4 * q_i)
            q[k_index + 1] = (R[k_index, i_index] + R[i_index, k_index]) / (4 * q_i)

        return cls.from_quaternion(q)

    @classmethod
    def from_basis(cls, t1, t2):
        """Create the object from two basis vectors t1, t2.

        t2 will be orthogonalized on t1, and t3 will be calculated with
        the cross product.
        """

        t1_norm = _np.linalg.norm(t1)
        if t1_norm < _bme.eps_quaternion:
            raise ValueError(f"The given vector t1 can not be a zero vector, got {t1}.")
        t1_normal = t1 / t1_norm

        t2_ortho = t2 - t1_normal * _np.dot(t1_normal, t2)
        t2_ortho_norm = _np.linalg.norm(t2_ortho)
        if t2_ortho_norm < _bme.eps_quaternion:
            raise ValueError(
                f"Got two vectors {t1} and {t2} that are not linear independent."
            )
        t2_normal = t2_ortho / t2_ortho_norm
        t3_normal = _np.cross(t1_normal, t2_normal)

        R = _np.transpose([t1_normal, t2_normal, t3_normal])
        return cls.from_rotation_matrix(R)

    @classmethod
    def from_rotation_vector(cls, rotation_vector):
        """Create the object from a rotation vector."""

        q = _np.zeros(4)
        rotation_vector = _np.asarray(rotation_vector)
        phi = _np.linalg.norm(rotation_vector)
        q[0] = _np.cos(0.5 * phi)
        if phi < _bme.eps_quaternion:
            # This is the Taylor series expansion of sin(phi/2)/phi around phi=0
            q[1:] = 0.5 * rotation_vector
        else:
            q[1:] = _np.sin(0.5 * phi) / phi * rotation_vector
        return cls.from_quaternion(q)

    def check(self):
        """Perform all checks for the rotation."""
        self.check_uniqueness()
        self.check_quaternion_constraint()

    def check_uniqueness(self):
        """We always want q0 to be positive -> the range for the rotational
        angle is 0 <= phi <= pi."""

        if self.q[0] < 0:
            self.q *= -1

    def check_quaternion_constraint(self):
        """We want to check that q.q = 1."""

        if _np.abs(1 - _np.linalg.norm(self.q)) > _bme.eps_quaternion:
            raise ValueError(
                f"The rotation object is corrupted. q.q does not equal 1! q={self.q}"
            )

    def get_rotation_matrix(self):
        """Return the rotation matrix for this rotation.

        (Krenk (3.50))
        """
        q_skew = skew_matrix(self.q[1:])
        R = (
            (self.q[0] ** 2 - _np.dot(self.q[1:], self.q[1:])) * _np.eye(3)
            + 2 * self.q[0] * q_skew
            + 2 * ([self.q[1:]] * _np.transpose([self.q[1:]]))
        )

        return R

    def get_quaternion(self):
        """Return the quaternion for this rotation, as numpy array (copy)."""
        return _np.array(self.q)

    def get_numpy_quaternion(self):
        """Return a numpy quaternion object representing this rotation
        (copy)."""
        return _quaternion.from_float_array(_np.array(self.q))

    def get_rotation_vector(self):
        """Return the rotation vector for this object."""

        self.check()

        norm = _np.linalg.norm(self.q[1:])
        phi = 2 * _np.arctan2(norm, self.q[0])

        if phi < _bme.eps_quaternion:
            # For small angles return the Taylor series expansion of phi/sin(phi/2)
            scale_factor = 2
        else:
            scale_factor = phi / _np.sin(phi / 2)
            if _np.abs(_np.abs(phi) - _np.pi) < _bme.eps_quaternion:
                # For rotations of exactly +-pi, numerical issues might occur, resulting in
                # a rotation vector that is non-deterministic. The result is correct, but
                # the sign can switch due to different implementation of basic underlying
                # math functions. This is especially triggered when using this code with
                # different OS. To avoid this, we scale the rotation axis in such a way,
                # for a rotation angle of +-pi, the first component of the rotation axis
                # that is not 0 is positive.
                for i_dir in range(3):
                    if _np.abs(self.q[1 + i_dir]) > _bme.eps_quaternion:
                        if self.q[1 + i_dir] < 0:
                            scale_factor *= -1
                        break
        return self.q[1:] * scale_factor

    def get_transformation_matrix(self):
        """Return the transformation matrix for this rotation.

        The transformation matrix maps the (infinitesimal)
        multiplicative rotational increments onto the additive ones.
        """

        omega = self.get_rotation_vector()
        omega_norm = _np.linalg.norm(omega)

        # We have to take the inverse of the the rotation angle here, therefore,
        # we have a branch for small angles where the singularity is not present.
        if omega_norm**2 > _bme.eps_quaternion:
            # Taken from Jelenic and Crisfield (1999) Equation (2.5)
            omega_dir = omega / omega_norm
            omega_skew = skew_matrix(omega)
            transformation_matrix = (
                _np.outer(omega_dir, omega_dir)
                - 0.5 * omega_skew
                + 0.5
                * omega_norm
                / _np.tan(0.5 * omega_norm)
                * (_np.identity(3) - _np.outer(omega_dir, omega_dir))
            )
        else:
            # This is the constant part of the Taylor series expansion. If this
            # function is used with automatic differentiation, higher order
            # terms have to be added!
            transformation_matrix = _np.identity(3)
        return transformation_matrix

    def get_transformation_matrix_inv(self):
        """Return the inverse of the transformation matrix for this rotation.

        The inverse of the transformation matrix maps the
        (infinitesimal) additive rotational increments onto the
        multiplicative ones.
        """

        omega = self.get_rotation_vector()
        omega_norm = _np.linalg.norm(omega)

        # We have to take the inverse of the the rotation angle here, therefore,
        # we have a branch for small angles where the singularity is not present.
        if omega_norm**2 > _bme.eps_quaternion:
            # Taken from Jelenic and Crisfield (1999) Equation (2.5)
            omega_dir = omega / omega_norm
            omega_skew = skew_matrix(omega)
            transformation_matrix_inverse = (
                (1.0 - _np.sin(omega_norm) / omega_norm)
                * _np.outer(omega_dir, omega_dir)
                + _np.sin(omega_norm) / omega_norm * _np.identity(3)
                + (1.0 - _np.cos(omega_norm)) / omega_norm**2 * omega_skew
            )
        else:
            # This is the constant part of the Taylor series expansion. If this
            # function is used with automatic differentiation, higher order
            # terms have to be added!
            transformation_matrix_inverse = _np.identity(3)
        return transformation_matrix_inverse

    def inv(self):
        """Return the inverse of this rotation."""

        tmp_quaternion = self.q.copy()
        tmp_quaternion[0] *= -1.0
        return Rotation.from_quaternion(tmp_quaternion)

    def __mul__(self, other):
        """Add this rotation to another, or apply it on a vector."""

        # Check if the other object is also a rotation.
        if isinstance(other, Rotation):
            # Get quaternions of the two objects.
            p = self.q
            q = other.q
            # Add the rotations.
            added_rotation = _np.zeros_like(self.q)
            added_rotation[0] = p[0] * q[0] - _np.dot(p[1:], q[1:])
            added_rotation[1:] = p[0] * q[1:] + q[0] * p[1:] + _np.cross(p[1:], q[1:])
            return Rotation.from_quaternion(added_rotation)
        elif isinstance(other, (list, _np.ndarray)) and len(other) == 3:
            # Apply rotation to vector.
            return _np.dot(self.get_rotation_matrix(), _np.asarray(other))
        raise NotImplementedError("Error, not implemented, does not make sense anyway!")

    def __eq__(self, other):
        """Check if the other rotation is equal to this one."""

        if isinstance(other, Rotation):
            return bool(
                (_np.linalg.norm(self.q - other.q) < _bme.eps_quaternion)
                or (_np.linalg.norm(self.q + other.q) < _bme.eps_quaternion)
            )
        else:
            return object.__eq__(self, other)

    def copy(self):
        """Return a copy of this object."""
        return Rotation.from_quaternion(self.q, normalized=True)

    def __str__(self):
        """String representation of object."""

        self.check()
        return f"Rotation:\n    q0: {self.q[0]}\n    q: {self.q[1:]}"


def add_rotations(
    rotation_21: Rotation | _NDArray[_quaternion.quaternion],
    rotation_10: Rotation | _NDArray[_quaternion.quaternion],
) -> _NDArray[_quaternion.quaternion]:
    """Multiply rotations onto another.

    Args:
        rotation_10: The first rotation(s) that are applied.
        rotation_21: The second rotation(s) that are applied.

    Returns:
        An array with the compound quaternions.
    """

    # Transpose the arrays, to work with the following code.
    if isinstance(rotation_10, Rotation):
        rot1 = rotation_10.get_quaternion().transpose()
    else:
        rot1 = _np.transpose(rotation_10)
    if isinstance(rotation_21, Rotation):
        rot2 = rotation_21.get_quaternion().transpose()
    else:
        rot2 = _np.transpose(rotation_21)

    if rot1.size > rot2.size:
        rotnew = _np.zeros_like(rot1)
    else:
        rotnew = _np.zeros_like(rot2)

    # Multiply the two rotations (code is taken from /utility/rotation.nb).
    rotnew[0] = (
        rot1[0] * rot2[0] - rot1[1] * rot2[1] - rot1[2] * rot2[2] - rot1[3] * rot2[3]
    )
    rotnew[1] = (
        rot1[1] * rot2[0] + rot1[0] * rot2[1] + rot1[3] * rot2[2] - rot1[2] * rot2[3]
    )
    rotnew[2] = (
        rot1[2] * rot2[0] - rot1[3] * rot2[1] + rot1[0] * rot2[2] + rot1[1] * rot2[3]
    )
    rotnew[3] = (
        rot1[3] * rot2[0] + rot1[2] * rot2[1] - rot1[1] * rot2[2] + rot1[0] * rot2[3]
    )

    return rotnew.transpose()


def rotate_coordinates(
    coordinates: _NDArray,
    rotation: Rotation | _NDArray[_quaternion.quaternion],
    *,
    origin=None,
):
    """Rotate all given coordinates.

    Args:
        coordinates: Array of 3D coordinates to be rotated
        rotation: The rotation(s) that will be applied to the coordinates. If
            this is an array it has to hold a quaternion for each coordinate.
    origin (3D vector):  If this is given, the mesh is rotated about this
        point. Defaults to (0, 0, 0).
    """

    if isinstance(rotation, Rotation):
        rotation = rotation.get_quaternion().transpose()

    # Check if origin has to be added
    if origin is None:
        origin = [0.0, 0.0, 0.0]

    # New position array
    coordinates_new = _np.zeros_like(coordinates)

    # Evaluate the new positions using the numpy data structure
    # (code is taken from /utility/rotation.nb)
    rotation = rotation.transpose()

    q0_q0 = _np.square(rotation[0])
    q0_q1_2 = 2.0 * rotation[0] * rotation[1]
    q0_q2_2 = 2.0 * rotation[0] * rotation[2]
    q0_q3_2 = 2.0 * rotation[0] * rotation[3]

    q1_q1 = _np.square(rotation[1])
    q1_q2_2 = 2.0 * rotation[1] * rotation[2]
    q1_q3_2 = 2.0 * rotation[1] * rotation[3]

    q2_q2 = _np.square(rotation[2])
    q2_q3_2 = 2.0 * rotation[2] * rotation[3]

    q3_q3 = _np.square(rotation[3])

    coordinates_new[:, 0] = (
        (q0_q0 + q1_q1 - q2_q2 - q3_q3) * (coordinates[:, 0] - origin[0])
        + (q1_q2_2 - q0_q3_2) * (coordinates[:, 1] - origin[1])
        + (q0_q2_2 + q1_q3_2) * (coordinates[:, 2] - origin[2])
    )
    coordinates_new[:, 1] = (
        (q1_q2_2 + q0_q3_2) * (coordinates[:, 0] - origin[0])
        + (q0_q0 - q1_q1 + q2_q2 - q3_q3) * (coordinates[:, 1] - origin[1])
        + (-q0_q1_2 + q2_q3_2) * (coordinates[:, 2] - origin[2])
    )
    coordinates_new[:, 2] = (
        (-q0_q2_2 + q1_q3_2) * (coordinates[:, 0] - origin[0])
        + (q0_q1_2 + q2_q3_2) * (coordinates[:, 1] - origin[1])
        + (q0_q0 - q1_q1 - q2_q2 + q3_q3) * (coordinates[:, 2] - origin[2])
    )

    if origin is not None:
        coordinates_new += origin

    return coordinates_new


def smallest_rotation(q: Rotation, t):
    """Get the triad that results from the smallest rotation (rotation without
    twist) from the triad q such that the rotated first basis vector aligns
    with t. For more details see Christoph Meier's dissertation chapter 2.1.2.

    Args
    ----
    q: Rotation
        Starting triad.
    t: Vector in R3
        Direction of the first basis of the rotated triad.
    Return
    ----
    q_sr: Rotation
        The triad that results from a smallest rotation.
    """

    R_old = q.get_rotation_matrix()
    g1_old = R_old[:, 0]
    g1 = _np.asarray(t) / _np.linalg.norm(t)

    # Quaternion components of relative rotation
    q_rel = _np.zeros(4)

    # The scalar quaternion part is cos(alpha/2) this is equal to
    q_rel[0] = _np.linalg.norm(0.5 * (g1_old + g1))

    # Vector part of the quaternion is sin(alpha/2)*axis
    q_rel[1:] = _np.cross(g1_old, g1) / (2.0 * q_rel[0])

    return Rotation.from_quaternion(q_rel) * q


def get_rotation_vector_series(
    rotation_vectors: list | _NDArray | list[Rotation],
) -> _NDArray:
    """Return an array containing the rotation vectors representing the given
    rotation vectors.

    The main feature of this function is, that the returned rotation
    vectors don't have jumps when the rotation angle exceeds 2*pi. We
    return a "continuous" series of rotation vectors that can be used to
    interpolate between them.

    Note: Interpolating between the returned rotation vectors will in general
    result in a non-objective interpolation.

    Args:
        rotation_vectors: A list or array containing the rotation vectors, has
            to be in order.

    Returns:
        An array containing the "continuous" rotation vectors.
    """

    if isinstance(rotation_vectors, list):
        rotation_vector_array = _np.zeros((len(rotation_vectors), 3))
        for i_rotation, rotation_entry in enumerate(rotation_vectors):
            if isinstance(rotation_entry, Rotation):
                rotation_vector_array[i_rotation] = rotation_entry.get_rotation_vector()
            else:
                rotation_vector_array[i_rotation] = rotation_entry

    def closest_multiple_of_two_pi(x: float) -> float:
        """Given the value x, return the multiple of 2*pi that is closest to
        it."""
        return 2.0 * _np.pi * round(x / (2 * _np.pi))

    rotation_vectors_continuous = _np.zeros((len(rotation_vector_array), 3))
    rotation_vectors_continuous[0, :] = rotation_vector_array[0]

    for i, current_rotation_vector in enumerate(rotation_vector_array[1:]):
        rotation_vector_last = rotation_vector_array[i]
        theta = _np.linalg.norm(current_rotation_vector)

        if theta < _bme.eps_quaternion:
            # The current rotation is an identity rotation.
            theta_last = _np.linalg.norm(rotation_vector_last)
            if theta_last < _bme.eps_quaternion:
                # The last rotation was also an identity rotation, so simply add
                # an empty rotation vector
                rotation_vectors_continuous[i + 1] = [0.0, 0.0, 0.0]
            else:
                # The last rotation was not a zero rotation. In this case we simply
                # take the direction of the last rotation vector and set it to to the
                # closest length which is a multiple of 2*pi.
                rotation_vectors_continuous[i + 1] = (
                    rotation_vector_last
                    / theta_last
                    * closest_multiple_of_two_pi(theta_last)
                )
        else:
            axis = current_rotation_vector / theta
            # This is the offset from the current rotation vector to the previous one
            # in the sense of a multiple of 2*pi.
            multiple_of_two_pi = closest_multiple_of_two_pi(
                (_np.dot(axis, rotation_vector_last) - theta)
            )
            rotation_vectors_continuous[i + 1] = (multiple_of_two_pi + theta) * axis
    return rotation_vectors_continuous



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file implements materials for 4C beams and solids."""

import numpy as _np

from beamme.core.material import Material as _Material
from beamme.core.material import MaterialBeamBase as _MaterialBeamBase
from beamme.core.material import MaterialSolidBase as _MaterialSolidBase


def get_all_contained_materials(
    material: _Material, _visited_materials: set[int] | None = None
) -> list[_Material]:
    """Recursively collect all materials contained within a material, including
    nested ones.

    Args:
        material:
            The root material from which to collect contained materials.
        _visited_materials:
            Internal parameter used to track visited materials and prevent
            infinite recursion in case of circular references.
            Users should not pass this manually.

    Returns:
        A flat list containing the given material and all nested materials.

    Raises:
        ValueError:
            If a circular material reference is detected.
    """

    if _visited_materials is None:
        _visited_materials = set()

    material_id = id(material)
    if material_id in _visited_materials:
        raise ValueError("Circular material reference detected!")
    _visited_materials.add(material_id)

    contained_materials = [material]

    if "MATIDS" in material.data:
        for item in material.data["MATIDS"]:
            if isinstance(item, _Material):
                contained_materials.extend(
                    get_all_contained_materials(item, _visited_materials)
                )

    return contained_materials


class MaterialReissner(_MaterialBeamBase):
    """Holds material definition for Reissner beams."""

    def __init__(
        self,
        shear_correction=1.0,
        *,
        by_modes=False,
        scale_axial_rigidity=1.0,
        scale_shear_rigidity=1.0,
        scale_torsional_rigidity=1.0,
        scale_bending_rigidity=1.0,
        **kwargs,
    ):
        if by_modes:
            mat_string = "MAT_BeamReissnerElastHyper_ByModes"
        else:
            mat_string = "MAT_BeamReissnerElastHyper"

        super().__init__(material_string=mat_string, **kwargs)

        # Shear factor for Reissner beam.
        self.shear_correction = shear_correction

        self.by_modes = by_modes

        # Scaling factors to influence a single stiffness independently
        self.scale_axial_rigidity = scale_axial_rigidity
        self.scale_shear_rigidity = scale_shear_rigidity
        self.scale_torsional_rigidity = scale_torsional_rigidity
        self.scale_bending_rigidity = scale_bending_rigidity

        if not by_modes and not all(
            _np.isclose(x, 1.0)
            for x in (
                scale_axial_rigidity,
                scale_shear_rigidity,
                scale_torsional_rigidity,
                scale_bending_rigidity,
            )
        ):
            raise ValueError(
                "Scaling factors are only supported for MAT_BeamReissnerElastHyper_ByModes"
            )

    def dump_to_list(self):
        """Return a list with the (single) item representing this material."""

        if self.radius is None or self.youngs_modulus is None:
            raise ValueError(
                "Radius and Young's modulus must be provided for beam materials."
            )

        if (
            self.area is None
            and self.mom2 is None
            and self.mom3 is None
            and self.polar is None
        ):
            area, mom2, mom3, polar = self.calc_area_stiffness()
        elif (
            self.area is not None
            and self.mom2 is not None
            and self.mom3 is not None
            and self.polar is not None
        ):
            area = self.area
            mom2 = self.mom2
            mom3 = self.mom3
            polar = self.polar
        else:
            raise ValueError(
                "Either all relevant material parameters are set "
                "by the user, or a circular cross-section will be assumed. "
                "A combination is not possible"
            )

        if self.by_modes:
            shear_modulus = self.youngs_modulus / (2.0 * (1.0 + self.nu))

            data = {
                "EA": (self.youngs_modulus * area) * self.scale_axial_rigidity,
                "GA2": (shear_modulus * area * self.shear_correction)
                * self.scale_shear_rigidity,
                "GA3": (shear_modulus * area * self.shear_correction)
                * self.scale_shear_rigidity,
                "GI_T": (shear_modulus * polar) * self.scale_torsional_rigidity,
                "EI2": (self.youngs_modulus * mom2) * self.scale_bending_rigidity,
                "EI3": (self.youngs_modulus * mom3) * self.scale_bending_rigidity,
                "RhoA": self.density * area,
                "MASSMOMINPOL": self.density * (mom2 + mom3),
                "MASSMOMIN2": self.density * mom2,
                "MASSMOMIN3": self.density * mom3,
            }

        else:
            data = {
                "YOUNG": self.youngs_modulus,
                "POISSONRATIO": self.nu,
                "DENS": self.density,
                "CROSSAREA": area,
                "SHEARCORR": self.shear_correction,
                "MOMINPOL": polar,
                "MOMIN2": mom2,
                "MOMIN3": mom3,
            }

        if self.interaction_radius is not None:
            data["INTERACTIONRADIUS"] = self.interaction_radius

        return {"MAT": self.i_global + 1, self.material_string: data}


class MaterialReissnerElastoplastic(MaterialReissner):
    """Holds elasto-plastic material definition for Reissner beams."""

    def __init__(
        self,
        *,
        yield_moment=None,
        isohardening_modulus_moment=None,
        torsion_plasticity=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.material_string = "MAT_BeamReissnerElastPlastic"

        if yield_moment is None or isohardening_modulus_moment is None:
            raise ValueError(
                "The yield moment and the isohardening modulus for moments must be specified "
                "for plasticity."
            )

        self.yield_moment = yield_moment
        self.isohardening_modulus_moment = isohardening_modulus_moment
        self.torsion_plasticity = torsion_plasticity

    def dump_to_list(self):
        """Return a list with the (single) item representing this material."""
        super_list = super().dump_to_list()
        mat_dict = super_list[self.material_string]
        mat_dict["YIELDM"] = self.yield_moment
        mat_dict["ISOHARDM"] = self.isohardening_modulus_moment
        mat_dict["TORSIONPLAST"] = self.torsion_plasticity
        return super_list


class MaterialKirchhoff(_MaterialBeamBase):
    """Holds material definition for Kirchhoff beams."""

    def __init__(self, is_fad=False, **kwargs):
        super().__init__(material_string="MAT_BeamKirchhoffElastHyper", **kwargs)
        self.is_fad = is_fad

    def dump_to_list(self):
        """Return a list with the (single) item representing this material."""

        if self.radius is None or self.youngs_modulus is None:
            raise ValueError(
                "Radius and Young's modulus must be provided for beam materials."
            )

        if (
            self.area is None
            and self.mom2 is None
            and self.mom3 is None
            and self.polar is None
        ):
            area, mom2, mom3, polar = self.calc_area_stiffness()
        elif (
            self.area is not None
            and self.mom2 is not None
            and self.mom3 is not None
            and self.polar is not None
        ):
            area = self.area
            mom2 = self.mom2
            mom3 = self.mom3
            polar = self.polar
        else:
            raise ValueError(
                "Either all relevant material parameters are set "
                "by the user, or a circular cross-section will be assumed. "
                "A combination is not possible"
            )
        data = {
            "YOUNG": self.youngs_modulus,
            "SHEARMOD": self.youngs_modulus / (2.0 * (1.0 + self.nu)),
            "DENS": self.density,
            "CROSSAREA": area,
            "MOMINPOL": polar,
            "MOMIN2": mom2,
            "MOMIN3": mom3,
            "FAD": self.is_fad,
        }
        if self.interaction_radius is not None:
            data["INTERACTIONRADIUS"] = self.interaction_radius
        return {"MAT": self.i_global + 1, self.material_string: data}


class MaterialEulerBernoulli(_MaterialBeamBase):
    """Holds material definition for Euler Bernoulli beams."""

    def __init__(self, **kwargs):
        super().__init__(
            material_string="MAT_BeamKirchhoffTorsionFreeElastHyper", **kwargs
        )

    def dump_to_list(self):
        """Return a list with the (single) item representing this material."""

        if self.radius is None or self.youngs_modulus is None:
            raise ValueError(
                "Radius and Young's modulus must be provided for beam materials."
            )

        area, mom2, _, _ = self.calc_area_stiffness()
        if self.area is None and self.mom2 is None:
            area, mom2, _, _ = self.calc_area_stiffness()
        elif self.area is not None and self.mom2 is not None:
            area = self.area
            mom2 = self.mom2
        else:
            raise ValueError(
                "Either all relevant material parameters are set "
                "by the user, or a circular cross-section will be assumed. "
                "A combination is not possible"
            )
        data = {
            "YOUNG": self.youngs_modulus,
            "DENS": self.density,
            "CROSSAREA": area,
            "MOMIN": mom2,
        }
        return {"MAT": self.i_global + 1, self.material_string: data}


class MaterialSolid(_MaterialSolidBase):
    """Base class for a material for solids."""

    def __init__(self, material_string=None, **kwargs):
        """Set the material values for a solid."""
        self.material_string = material_string
        super().__init__(**kwargs)

    def dump_to_list(self):
        """Return a list with the (single) item representing this material."""

        return {"MAT": self.i_global + 1, self.material_string: self.data}


class MaterialStVenantKirchhoff(MaterialSolid):
    """Holds material definition for StVenant Kirchhoff solids."""

    def __init__(self, youngs_modulus=None, nu=None, density=None):
        if youngs_modulus is None or nu is None:
            raise ValueError(
                "Young's modulus and Poisson's ratio must be provided "
                "for StVenant Kirchhoff solid materials."
            )
        data = {"YOUNG": youngs_modulus, "NUE": nu}
        if density is not None:
            data["DENS"] = density
        super().__init__(
            material_string="MAT_Struct_StVenantKirchhoff",
            data=data,
        )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test the functionality of the Cosserat curve module."""

import numpy as np
import pytest
import pyvista as pv
import quaternion

from beamme.core.element_beam import Beam3
from beamme.core.material import MaterialBeamBase
from beamme.core.rotation import Rotation
from beamme.cosserat_curve.cosserat_curve import CosseratCurve
from beamme.cosserat_curve.warping_along_cosserat_curve import (
    create_transform_boundary_conditions,
    get_mesh_transformation,
    warp_mesh_along_curve,
)
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.model_importer import import_four_c_model
from beamme.mesh_creation_functions.beam_helix import create_beam_mesh_helix


def load_cosserat_curve_from_file(get_corresponding_reference_file_path, **kwargs):
    """Load the centerline coordinates from the reference files and create the
    Cosserat curve."""
    point_coordinates_file = get_corresponding_reference_file_path(
        reference_file_base_name="test_cosserat_curve_centerline", extension="txt"
    )
    coordinates = np.loadtxt(
        point_coordinates_file, comments="#", delimiter=",", unpack=False
    )
    return CosseratCurve(coordinates, **kwargs)


def create_beam_solid_input_file(
    beam_type, material, get_corresponding_reference_file_path
):
    """Create a beam and solid input file for testing purposes."""

    _, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_cosserat_curve_mesh"
        ),
        convert_input_to_mesh=True,
    )

    create_beam_mesh_helix(
        mesh,
        beam_type,
        material,
        [0, 0, 1],
        [0, 0, 0],
        [2, 0, 0],
        helix_angle=0.4,
        turns=3,
        n_el=5,
    )
    return mesh


@pytest.mark.parametrize(
    ("twist_type", "twist_angle", "starting_triad_guess"),
    [
        (None, 0.0, None),
        ("angle", np.pi / 5, None),
        ("starting_rotation", 0.3293917321011114, Rotation([-0.5, 3, -0.5], 2)),
    ],
)
def test_cosserat_curve_translate_and_rotate(
    twist_type,
    twist_angle,
    starting_triad_guess,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test that a curve can be loaded, rotated and transformed."""

    if twist_type is None:
        curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)
    elif twist_type == "angle":
        curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)
        curve.twist(twist_angle)
    elif starting_triad_guess is not None:
        curve = load_cosserat_curve_from_file(
            get_corresponding_reference_file_path,
            starting_triad_guess=starting_triad_guess,
        )
    relative_rotation = Rotation([1, 0, 0], twist_angle)

    # Translate the curve so that the start is at the origin
    curve.translate(-curve.centerline_interpolation(5.0))

    # Rotate the curve around its center point
    origin = [1, 2, 3]
    rotation = Rotation([1, 2, 3], np.pi / 7)
    curve.rotate(rotation, origin=origin)

    # Get the points and rotations at certain points
    t = list(map(float, range(-10, 30, 5)))
    sol_half_pos, sol_half_q = curve.get_centerline_positions_and_rotations(
        t, factor=0.5
    )
    sol_full_pos, sol_full_q = curve.get_centerline_positions_and_rotations(
        t, factor=1.0
    )

    def load_compare(name):
        """Load the compare files and return a numpy array."""
        return np.loadtxt(
            get_corresponding_reference_file_path(
                additional_identifier=name, extension="txt"
            )
        )

    def get_compare_rot_with_twist(name):
        """Apply twist rotations to reference files."""
        rotations = (
            quaternion.from_float_array(load_compare(name))
            * relative_rotation.get_numpy_quaternion()
        )
        return rotations

    # Note: The tolerances are increased due to issues with the spline
    #     interpolation, see `beamme.cosserat_curve.get_spline_interpolation`.
    assert_results_close(sol_half_pos, load_compare("pos_half_ref"), atol=1e-12)
    assert_results_close(sol_half_q, get_compare_rot_with_twist("q_half_ref"))
    assert_results_close(sol_full_pos, load_compare("pos_full_ref"), atol=1e-12)
    assert_results_close(sol_full_q, get_compare_rot_with_twist("q_full_ref"))


def test_cosserat_curve_bad_guess_triad(get_corresponding_reference_file_path):
    """Check that an error is thrown for a bad guess triad."""
    with pytest.raises(ValueError):
        load_cosserat_curve_from_file(
            get_corresponding_reference_file_path,
            starting_triad_guess=Rotation([-0.5, 3, -0.5], 2)
            * Rotation([0, 0, 1], np.pi * 0.5),
        )


def test_cosserat_curve_vtk_representation(
    tmp_path, get_corresponding_reference_file_path, assert_results_close
):
    """Test the vtk representation of the Cosserat curve."""

    reference_path = get_corresponding_reference_file_path(extension="vtu")
    result_path = tmp_path / reference_path.name

    curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)
    pv.UnstructuredGrid(curve.get_pyvista_polyline()).save(result_path)

    assert_results_close(reference_path, result_path)


@pytest.mark.parametrize(
    ("factors", "n_steps", "reference_name"),
    [([0.1, 0.7], None, "factors"), (None, 3, "steps")],
)
def test_cosserat_curve_pvd_series(
    factors,
    n_steps,
    reference_name,
    tmp_path,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test the pvd series representation of the Cosserat curve."""

    reference_path = get_corresponding_reference_file_path(
        additional_identifier=reference_name, extension="pvd"
    )
    result_path = tmp_path / reference_path.name

    curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)
    curve.write_pvd_series(result_path, factors=factors, n_steps=n_steps)

    assert_results_close(reference_path, result_path)
    for i_step in range(len(factors) if factors is not None else n_steps):
        step_reference_path = get_corresponding_reference_file_path(
            additional_identifier=reference_name + "." + str(i_step),
            extension="vtu",
        )
        step_result_path = tmp_path / step_reference_path.name
        assert_results_close(step_reference_path, step_result_path)


def test_cosserat_curve_pvd_series_arguments(
    get_corresponding_reference_file_path, assert_results_close
):
    """Test the that arguments are correctly processed in the pvd series
    representation of the Cosserat curve."""

    curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)

    pvd_name = "temp.pvd"

    # Check that errors are raised for invalid argument combinations
    with pytest.raises(
        ValueError, match="The output path must have a .pvd suffix, got .vtu"
    ):
        curve.write_pvd_series("temp.vtu")

    with pytest.raises(
        ValueError,
        match="The keyword arguments 'factors' and 'n_steps' are mutually exclusive.",
    ):
        curve.write_pvd_series(pvd_name, factors=[0.0, 1.0], n_steps=2)

    with pytest.raises(
        ValueError,
        match="One of the keyword arguments 'factors' or 'n_steps' must be provided.",
    ):
        curve.write_pvd_series(pvd_name)


def test_cosserat_curve_project_point(
    get_corresponding_reference_file_path, assert_results_close
):
    """Test that the project point function works as expected."""

    # Load the curve
    curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)

    # Translate the curve so that the start is at the origin
    curve.translate(-curve.centerline_interpolation(0.0))

    # Check the projection results
    t_ref = 4.264045157204052
    assert_results_close(t_ref, curve.project_point([-5, 1, 1]))
    assert_results_close(t_ref, curve.project_point([-5, 1, 1], t0=2.0))
    assert_results_close(t_ref, curve.project_point([-5, 1, 1], t0=4.0))


def test_cosserat_curve_mesh_transformation(
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test that the get_mesh_transformation function works as expected."""

    curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)
    pos, rot = curve.get_centerline_position_and_rotation(0)
    rot = Rotation.from_quaternion(rot)
    curve.translate(-pos)
    curve.translate([1, 2, 3])

    mesh = create_beam_solid_input_file(
        Beam3, MaterialBeamBase(), get_corresponding_reference_file_path
    )
    pos, rot = get_mesh_transformation(
        curve,
        mesh.nodes,
        origin=[0.5, 1.0, -1.0],
        reference_rotation=(
            Rotation([0, 0, 1], -0.5 * np.pi) * Rotation([0, 1, 0], -0.5 * np.pi)
        ),
        n_steps=3,
    )

    assert_results_close(
        get_corresponding_reference_file_path(extension="json"),
        {"pos": pos, "rot": quaternion.as_float_array(rot)},
    )


def test_cosserat_curve_mesh_warp(
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Warp a balloon along a centerline."""

    # Load the curve
    curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)
    pos, rot = curve.get_centerline_position_and_rotation(0)
    rot = Rotation.from_quaternion(rot)
    curve.translate(-pos)
    curve.translate([1, 2, 3])

    # Warp the mesh. The reference coordinate system is rotated such that z axis is the longitudinal direction,
    # and x and y are the first and second cross-section basis vectors respectively.
    mesh = create_beam_solid_input_file(
        Beam3rHerm2Line3,
        get_default_test_beam_material(material_type="reissner"),
        get_corresponding_reference_file_path,
    )
    warp_mesh_along_curve(
        mesh,
        curve,
        origin=[0.5, 1.0, -1.0],
        reference_rotation=(
            Rotation([0, 0, 1], -0.5 * np.pi) * Rotation([0, 1, 0], -0.5 * np.pi)
        ),
    )
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_cosserat_curve_mesh_warp_transform_boundary_conditions(
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test the transform boundary creation function."""

    # Load the curve
    curve = load_cosserat_curve_from_file(get_corresponding_reference_file_path)
    pos, rot = curve.get_centerline_position_and_rotation(0)
    rot = Rotation.from_quaternion(rot)
    curve.translate(-pos)
    curve.translate([1, 2, 3])

    # Load the mesh
    mesh = create_beam_solid_input_file(
        Beam3rHerm2Line3,
        get_default_test_beam_material(material_type="reissner"),
        get_corresponding_reference_file_path,
    )

    # Apply the transform boundary conditions
    create_transform_boundary_conditions(
        mesh,
        curve,
        n_steps=3,
        origin=[2, 3, 0.5],
        reference_rotation=(
            Rotation([0, 0, 1], -0.5 * np.pi) * Rotation([0, 1, 0], -0.5 * np.pi)
        ),
    )

    # Note: The tolerances are increased due to issues with the spline
    #     interpolation, see `beamme.cosserat_curve.get_spline_interpolation`.
    assert_results_close(
        get_corresponding_reference_file_path(), mesh, rtol=1e-12, atol=1e-12
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the functionality of the function_utility
module."""

from beamme.four_c.function_utility import (
    create_linear_interpolation_dict,
    create_linear_interpolation_function,
)


def test_beamme_four_c_function_utility_linear_interpolation_function(
    assert_results_close,
):
    """Test that linear interpolation functions are created correctly."""

    t = [1.5, 2.5, 3.5, 10.0]
    values = [1.0, -1.0, 3.5, -10.3]

    fun = create_linear_interpolation_function(t, values)
    assert_results_close(
        [
            {"SYMBOLIC_FUNCTION_OF_SPACE_TIME": "var"},
            {
                "VARIABLE": 0,
                "NAME": "var",
                "TYPE": "linearinterpolation",
                "NUMPOINTS": 6,
                "TIMES": [-1000.0, 1.5, 2.5, 3.5, 10.0, 1010.0],
                "VALUES": [1.0, 1.0, -1.0, 3.5, -10.3, -10.3],
            },
        ],
        fun.data,
    )

    fun = create_linear_interpolation_function(t, values, function_type="My type")
    assert_results_close(
        [
            {"My type": "var"},
            {
                "VARIABLE": 0,
                "NAME": "var",
                "TYPE": "linearinterpolation",
                "NUMPOINTS": 6,
                "TIMES": [-1000.0, 1.5, 2.5, 3.5, 10.0, 1010.0],
                "VALUES": [1.0, 1.0, -1.0, 3.5, -10.3, -10.3],
            },
        ],
        fun.data,
    )

    function_definition = create_linear_interpolation_dict(
        t, values, variable_name="test", variable_index=12
    )
    assert_results_close(
        {
            "VARIABLE": 12,
            "NAME": "test",
            "TYPE": "linearinterpolation",
            "NUMPOINTS": 6,
            "TIMES": [-1000.0, 1.5, 2.5, 3.5, 10.0, 1010.0],
            "VALUES": [1.0, 1.0, -1.0, 3.5, -10.3, -10.3],
        },
        function_definition,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to unittest the functionality of the Rotation class."""

import numpy as np
import pytest

from beamme.core.conf import bme
from beamme.core.rotation import Rotation, get_rotation_vector_series, smallest_rotation


def get_rotation_matrix(axis, alpha):
    """Create a rotation about one of the Cartesian axis.

    Args
    ----
    axis: int
        0 - x
        1 - y
        2 - z
    angle: double rotation angle

    Return
    ----
    rot3D: array(3x3)
        Rotation matrix for this rotation
    """
    c, s = np.cos(alpha), np.sin(alpha)
    rot2D = np.array(((c, -s), (s, c)))
    index = [np.mod(j, 3) for j in range(axis, axis + 3) if not j == axis]
    rot3D = np.eye(3)
    rot3D[np.ix_(index, index)] = rot2D
    return rot3D


def test_beamme_core_rotations_cartesian_rotations(assert_results_close):
    """Create a rotation in all 3 directions.

    And compare with the rotation matrix.
    """

    theta = 1.0
    # Loop per direction.
    for i in range(3):
        rot3D = get_rotation_matrix(i, theta)
        axis = np.zeros(3)
        axis[i] = 1
        angle = theta
        rotation = Rotation(axis, angle)

        # Check if the rotation is the same if it is created from its own
        # quaternion and then created from its own rotation matrix.
        rotation = Rotation.from_quaternion(rotation.get_quaternion())
        rotation_matrix = Rotation.from_rotation_matrix(rotation.get_rotation_matrix())

        assert_results_close(rot3D, rotation_matrix.get_rotation_matrix())


def test_beamme_core_rotations_euler_angles(assert_results_close):
    """Create a rotation with Euler angles and compare to known results."""

    # Euler angles.
    alpha = 1.1
    beta = 1.2 * np.pi * 10
    gamma = -2.5

    # Create the rotation with rotation matrices.
    Rx = get_rotation_matrix(0, alpha)
    Ry = get_rotation_matrix(1, beta)
    Rz = get_rotation_matrix(2, gamma)
    R_euler = Rz.dot(Ry.dot(Rx))

    # Create the rotation with the Rotation object.
    rotation_x = Rotation([1, 0, 0], alpha)
    rotation_y = Rotation([0, 1, 0], beta)
    rotation_z = Rotation([0, 0, 1], gamma)
    rotation_euler = rotation_z * rotation_y * rotation_x
    assert_results_close(R_euler, rotation_euler.get_rotation_matrix())
    assert rotation_euler == Rotation.from_rotation_matrix(R_euler)

    # Direct formula for quaternions for Euler angles.
    quaternion = np.zeros(4)
    cy = np.cos(gamma * 0.5)
    sy = np.sin(gamma * 0.5)
    cr = np.cos(alpha * 0.5)
    sr = np.sin(alpha * 0.5)
    cp = np.cos(beta * 0.5)
    sp = np.sin(beta * 0.5)
    quaternion[0] = cy * cr * cp + sy * sr * sp
    quaternion[1] = cy * sr * cp - sy * cr * sp
    quaternion[2] = cy * cr * sp + sy * sr * cp
    quaternion[3] = sy * cr * cp - cy * sr * sp
    assert Rotation.from_quaternion(quaternion) == rotation_euler
    assert Rotation.from_quaternion(quaternion) == Rotation.from_quaternion(
        rotation_euler.get_quaternion()
    )
    assert Rotation.from_quaternion(quaternion) == Rotation.from_rotation_matrix(
        R_euler
    )


def test_beamme_core_rotations_negative_angles():
    """Check if a rotation is created correctly if a negative angle or a large
    angle is given."""

    vector = 10 * np.array([-1.234243, -2.334343, -1.123123])
    phi = -12.152101868665
    rot = Rotation(vector, phi)
    for i in range(2):
        assert rot == Rotation(vector, phi + 2 * i * np.pi)

    rot = Rotation.from_rotation_vector(vector)
    q = rot.q
    assert rot == Rotation.from_quaternion(-q)
    assert Rotation.from_quaternion(q) == Rotation.from_quaternion(-q)


def test_beamme_core_rotations_inverse_rotation():
    """Test the inv() function for rotations."""

    # Define test rotation.
    rot = Rotation([1, 2, 3], 2)

    # Check if inverse rotation gets identity rotation. Use two different
    # constructors for identity rotation.
    assert Rotation.from_rotation_vector([0, 0, 0]) == rot * rot.inv()
    assert Rotation() == rot * rot.inv()

    # Check that there is no warning or error when getting the vector for
    # an identity rotation.
    (rot * rot.inv()).get_rotation_vector()


def test_beamme_core_rotations_rotation_vector(assert_results_close):
    """Test if the rotation vector functions give a correct result."""

    # Calculate rotation vector and quaternion.
    axis = np.array([1.36568, -2.96784, 3.23346878])
    angle = 0.7189467
    rotation_vector = angle * axis / np.linalg.norm(axis)
    q = np.zeros(4)
    q[0] = np.cos(angle / 2)
    q[1:] = np.sin(angle / 2) * axis / np.linalg.norm(axis)

    # Check that the rotation object from the quaternion and rotation
    # vector are equal.
    rotation_from_vec = Rotation.from_rotation_vector(rotation_vector)
    assert Rotation.from_quaternion(q) == rotation_from_vec
    assert Rotation(axis, angle) == rotation_from_vec

    # Check that the same rotation vector is returned after being converted
    # to a quaternion.
    assert_results_close(rotation_vector, rotation_from_vec.get_rotation_vector())


def test_beamme_core_rotations_rotation_operator_overload(assert_results_close):
    """Test if the operator overloading gives a correct result."""

    # Calculate rotation and vector.
    axis = np.array([1.36568, -2.96784, 3.23346878])
    angle = 0.7189467
    rot = Rotation(axis, angle)
    vector = [2.234234, -4.213234, 6.345234]

    # Check the result of the operator overloading.
    result_vector = np.dot(rot.get_rotation_matrix(), vector)
    assert_results_close(result_vector, rot * vector)
    assert_results_close(result_vector, rot * np.array(vector))


@pytest.mark.parametrize(
    "vectors",
    [
        [[1, 0, 0], [0, -1, 0]],
        [[0, 0, 1], [0, 1, 0]],
        [[-1, 0, 0], [0, 1, 0]],
        [[0, 1, 0], [0, 0, 1]],
    ],
)
def test_beamme_core_rotations_from_rotation_matrix(vectors, assert_results_close):
    """Test if the correct quaternions are generated from a rotation matrix.

    The from_rotation_matrix function has different branches, the input
    vectors to this functions are chosen to trigger all branches.
    """

    t1, t2 = vectors
    rot = Rotation().from_basis(t1, t2)
    t1_rot = rot * [1, 0, 0]
    t2_rot = rot * [0, 1, 0]
    assert_results_close(t1, t1_rot)
    assert_results_close(t2, t2_rot)


def test_beamme_core_rotations_from_basis(assert_results_close):
    """Test the from_basis function for general input values."""

    t1 = [1, 2, 3]
    t2 = [-1, 3, 1]
    rot = Rotation().from_basis(t1, t2)
    t1_rot = rot * [1, 0, 0]
    t3_rot = rot * [0, 0, 1]

    # The first basis vector has to be in t1 direction
    assert_results_close(t1 / np.linalg.norm(t1), t1_rot)

    # The third basis vector has to be orthogonal to t1 and t2
    for vector in (t1, t2):
        assert_results_close(np.dot(vector, t3_rot), 0.0)

    # Check that an error is raised for wrong input parameters
    with pytest.raises(ValueError, match="The given vector t1"):
        Rotation().from_basis([0, 0, 0], [2, 4, 6])
    with pytest.raises(ValueError, match="Got two vectors"):
        Rotation().from_basis([1, 2, 3], [2, 4, 6])
    with pytest.raises(ValueError, match="Got two vectors"):
        Rotation().from_basis([1, 2, 3], [0, 0, 0])


def test_beamme_core_rotations_transformation_matrix(assert_results_close):
    """Test that the transformation matrix is computed correctly."""

    rotation_vector_large = [1.0, 2.0, np.pi / 5.0]
    rotation_large = Rotation.from_rotation_vector(rotation_vector_large)
    rotation_vector_small = (
        rotation_vector_large
        / np.linalg.norm(rotation_vector_large)
        / 10.0
        * bme.eps_quaternion
    )
    rotation_small = Rotation.from_rotation_vector(rotation_vector_small)

    # Test transformation matrix
    transformation_matrix_large_reference = np.array(
        [
            [0.5959488405656389, 0.49803685445056006, -0.9422331516950085],
            [-0.13028167626739845, 0.8717652242030102, 0.6155336966099826],
            [1.0577668483049911, -0.3844663033900173, 0.5403060272710478],
        ]
    )
    assert_results_close(
        rotation_large.get_transformation_matrix(),
        transformation_matrix_large_reference,
    )
    assert_results_close(rotation_small.get_transformation_matrix(), np.identity(3))

    # Test transformation matrix inverse
    transformation_matrix_inverse_large_reference = np.array(
        [
            [0.44154375784863675, 0.05812896596538626, 0.7037804689849043],
            [0.4501610612860693, 0.8227612782872283, -0.15228520755418612],
            [-0.5440964474342915, 0.4716532506554118, 0.36463746593568075],
        ]
    )
    assert_results_close(
        rotation_large.get_transformation_matrix_inv(),
        transformation_matrix_inverse_large_reference,
    )
    assert_results_close(rotation_small.get_transformation_matrix_inv(), np.identity(3))


def test_beamme_core_rotations_smallest_rotation_triad(assert_results_close):
    """Test that the smallest rotation triad is calculated correctly."""

    # Get the triad obtained by a smallest rotation from an arbitrary triad
    # onto an arbitrary tangent vector.
    rot = Rotation([1, 2, 3], 0.431 * np.pi)
    tan = [2.0, 3.0, -1.0]
    rot_smallest = smallest_rotation(rot, tan)

    rot_smallest_ref = [
        0.853329730651268,
        0.19771093216880734,
        0.25192421451158936,
        0.4114279380770031,
    ]
    assert_results_close(rot_smallest.q, rot_smallest_ref)


def test_beamme_core_rotations_error_accumulation_multiplication(assert_results_close):
    """Test that error accumulation of successive multiplications of rotations
    does not affect the results."""

    rotation_1 = Rotation([1, 2, 3], 0.3)
    rotation_2 = Rotation([1, -1, -2], np.pi / 6)
    rotation_3 = Rotation([-1, -2, -3], 7 * np.pi / 17)
    rotation = Rotation()
    for _ in range(100):
        rotation = rotation_1 * rotation * rotation_2
        rotation = rotation * rotation_3

    q_ref = [
        -0.38478914485223104,
        -0.0385171948379694,
        -0.49122781649072017,
        -0.780479962594468,
    ]
    assert_results_close(q_ref, rotation.q)


def test_beamme_core_rotations_error_accumulation_smallest_rotation(
    assert_results_close,
):
    """Test that error accumulation of successive smallest rotation mappings
    does not affect the results.

    Calculate the smallest rotation onto a vector and then rotate that
    vector "away" to calculate the next smallest rotation and so on...
    """

    tangent = [0.9, 0.1, -0.3]
    rotation_old = Rotation([1, 2, 3], 0.3)

    for _ in range(50):
        rotation_new = smallest_rotation(rotation_old, tangent)
        tangent = rotation_new * rotation_old.inv() * tangent
        rotation_old = rotation_new

    q_ref = [
        0.6329069205124062,
        0.13331392718187732,
        -0.5128773537467728,
        0.5644581887089211,
    ]
    assert_results_close(q_ref, rotation_new.q)


def test_beamme_core_rotations_rotations_rotation_vector_series(assert_results_close):
    """Test the function get_rotation_vector_series."""

    director = np.array([1, 2, 3])
    director = director / np.linalg.norm(director)
    director_2 = np.array([1.1, 2.5, 3.1])
    director_2 = 1.1 * director_2 / np.linalg.norm(director_2)

    # First rotation is 0, second is also zero but with a full 2pi rotation.
    rotation_vectors = [
        [0, 0, 0],
        2.0 * np.pi * director,
    ]
    rotation_vectors_ref = [[0, 0, 0], [0, 0, 0]]
    assert_results_close(
        get_rotation_vector_series(rotation_vectors), rotation_vectors_ref
    )

    # First rotation is close to 2pi, second is 0.
    rotation_vectors = [1.9 * np.pi * director, [0, 0, 0]]
    rotation_vectors_ref = [1.9 * np.pi * director, 2.0 * np.pi * director]
    assert_results_close(
        get_rotation_vector_series(rotation_vectors), rotation_vectors_ref
    )

    # Two similar rotations, no jump over 2pi.
    rotation_vectors = [director, director_2]
    rotation_vectors_ref = rotation_vectors
    assert_results_close(
        get_rotation_vector_series(rotation_vectors), rotation_vectors_ref
    )

    # Standard case, two similar rotations, without jump over 2pi
    rotation_vectors = [director, director_2]
    rotation_vectors_ref = rotation_vectors
    assert_results_close(
        get_rotation_vector_series(rotation_vectors), rotation_vectors_ref
    )

    # Standard case, two similar rotations, with jump over 2pi
    rotation_vectors = [
        director,
        (2 * np.pi + np.linalg.norm(director_2))
        * director_2
        / np.linalg.norm(director_2),
    ]
    rotation_vectors_ref = [director, director_2]
    assert_results_close(
        get_rotation_vector_series(rotation_vectors), rotation_vectors_ref
    )

    # Standard case, two similar rotations, with jump over 2pi, using rotation
    # objects.
    # With the rotation objects, there is no jump over 2pi, ut in this case we
    # expect the same result as for the previous case.
    # In theory it can be, that the rotation object leads to a different
    # "initial" rotation vector and then also a different series.
    rotation_vectors = [
        Rotation.from_rotation_vector(item) for item in rotation_vectors
    ]
    assert_results_close(
        get_rotation_vector_series(rotation_vectors), rotation_vectors_ref
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Integration tests for the 4C model importer of BeamMe."""

import re

import pytest

from beamme.core.mesh import Mesh
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.input_file import InputFile
from beamme.four_c.model_importer import (
    _extract_mesh_sections,
    import_cubitpy_model,
    import_four_c_model,
)
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from tests.create_test_models import create_tube_cubit


@pytest.mark.parametrize("full_import", (False, True))
@pytest.mark.cubitpy
def test_integration_four_c_model_importer_import_cubitpy_model(
    full_import, assert_results_close, get_corresponding_reference_file_path
):
    """Check that an import from a cubitpy object works as expected."""

    cubit = create_tube_cubit()
    input_file_cubit, mesh = import_cubitpy_model(
        cubit, convert_input_to_mesh=full_import
    )
    if full_import:
        input_file_cubit.add(mesh)

    assert_results_close(
        get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_tube"
        ),
        input_file_cubit,
    )


def test_integration_four_c_model_importer_import_nested_materials(
    get_default_test_solid_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Check if nested materials are imported correctly."""

    # Create a minimal solid input file.
    mesh = Mesh()
    material = get_default_test_solid_material(material_type="solid_nested")
    mesh.add(material)

    # Add the mesh to an input file, this converts the material to the FourCIPP format.
    input_file = InputFile()
    input_file.add(mesh)

    # Extract the mesh again. Only one material should be present in the extracted mesh.
    _, mesh_extracted = _extract_mesh_sections(input_file)
    assert len(mesh_extracted.materials) == 1

    # Check that the imported mesh can be added to a mesh that already contains a
    # material. This tests that the nested materials are correctly relinked during
    # the import.
    mesh_with_other_material = Mesh()
    mesh_with_other_material.add(
        get_default_test_solid_material(material_type="st_venant_kirchhoff")
    )
    input_file = InputFile()
    input_file.add(mesh_with_other_material)
    input_file.add(mesh_extracted)

    # Compare with reference file.
    assert_results_close(get_corresponding_reference_file_path(), input_file)


def test_integration_four_c_model_importer_import_nested_materials_error():
    """Check that an error is raised when importing nested materials with bad
    IDs."""

    # Create an input file with "bad" material IDs.
    input_file = InputFile()
    input_file["MATERIALS"] = [
        {"MAT": 1, "MAT_ElastHyper": {"NUMMAT": 2, "MATIDS": [2, 3], "DENS": 1.0}},
        {"MAT": 2, "ELAST_CoupSVK": {"YOUNG": 1.0, "NUE": 0.0}},
    ]

    # Try to extract the mesh, this should raise an error.
    with pytest.raises(
        KeyError,
        match=re.escape(
            "Material ID 3 not in material_id_map_all (available IDs: [1, 2])."
        ),
    ):
        _extract_mesh_sections(input_file)


@pytest.mark.parametrize("full_import", (False, True))
def test_integration_four_c_model_importer_non_consecutive_geometry_sets(
    full_import,
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test that we can import non-consecutively numbered geometry sets."""

    input_file, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            additional_identifier="input"
        ),
        convert_input_to_mesh=full_import,
    )

    material = get_default_test_beam_material(material_type="reissner")
    for i in range(3):
        beam_set = create_beam_mesh_line(
            mesh,
            Beam3rHerm2Line3,
            material,
            [i + 3, 0, 0],
            [i + 3, 0, 4],
            n_el=2,
        )
        mesh.add(beam_set)

    input_file.add(mesh)

    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier="full_import" if full_import else "dict_import"
        ),
        input_file,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file contains the wrapper for the LocSys condition for 4c."""

from typing import List as _List
from typing import Union as _Union

from beamme.core.boundary_condition import BoundaryCondition as _BoundaryCondition
from beamme.core.conf import bme as _bme
from beamme.core.function import Function as _Function
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.rotation import Rotation as _Rotation
from beamme.four_c.function_utility import (
    ensure_length_of_function_array as _ensure_length_of_function_array,
)


class LocSysCondition(_BoundaryCondition):
    """This object represents a locsys condition in 4C.

    It allows to rotate the local coordinate system used to apply
    Dirichlet boundary conditions.
    """

    def __init__(
        self,
        geometry_set: _GeometrySet,
        *,
        rotation: None | _Rotation = None,
        function_array: None | _List[_Union[_Function, int]] = None,
        update_node_position: bool = False,
        use_consistent_node_normal: bool = False,
        **kwargs,
    ):
        """Initialize the object.

        Args:
            geometry_set: Geometry that this boundary condition acts on
            rotation: Object that represents the rotation of the coordinate system.
            function_array:
                - If a single function is provided, it is used to scale the entire rotation.
                - If three functions are provided, they represent the components of a rotation vector,
                  in which case no explicit rotation should be passed.
            update_node_position: Flag to enable the updated node position
            use_consistent_node_normal: Flag to use a consistent node normal
        """

        # Check for invalid input arguments
        if (
            function_array is not None
            and len(function_array) > 1
            and rotation is not None
        ):
            raise ValueError(
                "If more than a single function is provided in `function_array`, "
                "no explicit `rotation` should be given. Either provide "
                "a rotation with a single function (scaling), or three "
                "functions (rotation vector components) without rotation."
            )

        # Validate provided function array
        if function_array is None:
            function_array = [0, 0, 0]
        else:
            function_array = _ensure_length_of_function_array(function_array, 3)

        # Validate provided rotation
        if rotation is None:
            rotation_vector = [1, 1, 1]
        else:
            rotation_vector = rotation.get_rotation_vector()

        condition_dict = {
            "ROTANGLE": rotation_vector,
            "FUNCT": function_array,
            "USEUPDATEDNODEPOS": int(update_node_position),
        }

        # Append the condition string with consistent normal type for line and surface geometry
        if (
            geometry_set.geometry_type is _bme.geo.line
            or geometry_set.geometry_type is _bme.geo.surface
        ):
            condition_dict["USECONSISTENTNODENORMAL"] = int(use_consistent_node_normal)
        elif use_consistent_node_normal:
            raise ValueError(
                "The keyword use_consistent_node_normal only works for line and surface geometries."
            )

        super().__init__(
            geometry_set, data=condition_dict, bc_type=_bme.bc.locsys, **kwargs
        )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Base testing framework infrastructure."""

import os
from pathlib import Path
from typing import Callable, Dict

import pytest
from _pytest.config import Config
from _pytest.config.argparsing import Parser

from beamme.core.conf import bme
from beamme.core.material import MaterialBeamBase
from beamme.four_c.material import (
    MaterialReissner,
    MaterialSolid,
    MaterialStVenantKirchhoff,
)

# Import additional confest files (split for better overview)
pytest_plugins = [
    "tests.conftest_performance_tests",
    "tests.conftest_result_comparison",
]


def pytest_addoption(parser: Parser) -> None:
    """Add custom command line options to pytest.

    Args:
        parser: Pytest parser
    """

    parser.addoption(
        "--4C",
        action="store_true",
        default=False,
        help="Execute standard and 4C based tests.",
    )

    parser.addoption(
        "--ArborX",
        action="store_true",
        default=False,
        help="Execute standard and ArborX based tests.",
    )

    parser.addoption(
        "--CubitPy",
        action="store_true",
        default=False,
        help="Execute standard and CubitPy based tests.",
    )

    parser.addoption(
        "--performance-tests",
        action="store_true",
        default=False,
        help="Execute standard and performance tests.",
    )

    parser.addoption(
        "--exclude-standard-tests",
        action="store_true",
        default=False,
        help="Exclude standard tests.",
    )


def pytest_collection_modifyitems(config: Config, items: list) -> None:
    """Filter tests based on their markers and provided command line options.

    Currently configured options:
        `pytest`: Execute standard tests with no markers
        `pytest --4C`: Execute standard tests and tests with the `fourc` marker
        `pytest --ArborX`: Execute standard tests and tests with the `arborx` marker
        `pytest --CubitPy`: Execute standard tests and tests with the `cubitpy` marker
        `pytest --performance-tests`: Execute standard tests and tests with the `performance` marker
        `pytest --exclude-standard-tests`: Execute tests with any other marker and exclude the standard unmarked tests

    Args:
        config: Pytest config
        items: Pytest list of tests
    """

    selected_tests = []

    # loop over all collected tests
    for item in items:
        # Get all set markers for current test (e.g. `fourc_arborx`, `cubitpy`, `performance`, ...)
        # We don't care about the "parametrize" marker here
        markers = [
            marker.name
            for marker in item.iter_markers()
            if not marker.name == "parametrize"
        ]

        for flag, marker in zip(
            ["--4C", "--ArborX", "--CubitPy", "--performance-tests"],
            ["fourc", "arborx", "cubitpy", "performance"],
        ):
            if config.getoption(flag) and marker in markers:
                selected_tests.append(item)

        if not markers and not config.getoption("--exclude-standard-tests"):
            selected_tests.append(item)

    deselected_tests = list(set(items) - set(selected_tests))

    items[:] = selected_tests
    config.hook.pytest_deselected(items=deselected_tests)


@pytest.fixture(autouse=True)
def run_before_each_test():
    """Reset the global bme object before each test."""
    bme.set_default_values()


@pytest.fixture(scope="session")
def reference_file_directory() -> Path:
    """Provide the path to the reference file directory.

    Returns:
        Path: A Path object representing the full path to the reference file directory.
    """

    testing_path = Path(__file__).resolve().parent
    return testing_path / "reference-files"


@pytest.fixture(scope="function")
def current_test_name(request: pytest.FixtureRequest) -> str:
    """Return the name of the current pytest test.

    Args:
        request: The pytest request object.

    Returns:
        str: The name of the current pytest test.
    """

    return request.node.originalname


@pytest.fixture(scope="function")
def get_corresponding_reference_file_path(
    reference_file_directory, current_test_name
) -> Callable:
    """Return function to get path to corresponding reference file for each
    test.

    Necessary to enable the function call through pytest fixtures.
    """

    def _get_corresponding_reference_file_path(
        reference_file_base_name: str | None = None,
        additional_identifier: str | None = None,
        additional_identifier_separator: str = "_",
        extension: str = "4C.yaml",
    ) -> Path:
        """Get path to corresponding reference file for each test. Also check
        if this file exists. Basename, additional identifier and extension can
        be adjusted.

        Args:
            reference_file_base_name: Basename of reference file, if none is
                provided the current test name is utilized
            additional_identifier: Additional identifier for reference file, by default none
            additional_identifier_separator: Separator between base name and
                additional identifier.
            extension: Extension of reference file, by default ".4C.yaml"

        Returns:
            Path to reference file.
        """

        corresponding_reference_file = reference_file_base_name or current_test_name

        if additional_identifier:
            corresponding_reference_file += (
                f"{additional_identifier_separator}{additional_identifier}"
            )

        corresponding_reference_file += "." + extension

        corresponding_reference_file_path = (
            reference_file_directory / corresponding_reference_file
        )

        if not os.path.isfile(corresponding_reference_file_path):
            raise AssertionError(
                f"File path: {corresponding_reference_file_path} does not exist"
            )

        return corresponding_reference_file_path

    return _get_corresponding_reference_file_path


@pytest.fixture(scope="function")
def get_bc_data() -> Callable:
    """Return a function to create a dummy definition for a boundary condition
    in 4C.

    Returns:
        A function to create a dummy boundary condition definition.
    """

    def _get_bc_data(*, identifier=None, num_dof: int = 3) -> Dict:
        """Return a dummy definition for a boundary condition in 4C that can be
        used for testing purposes.

        Args:
            identifier: Any value, will be written to the value for the first DOF. This can be used to create multiple boundary conditions and distinguish them in the input file.
            num_dof: Number of degrees of freedom constrained by this boundary condition.
        """

        val = [0] * num_dof
        if identifier is not None:
            val[0] = identifier

        return {
            "NUMDOF": num_dof,
            "ONOFF": [1] * num_dof,
            "VAL": val,
            "FUNCT": [0] * num_dof,
        }

    return _get_bc_data


@pytest.fixture(scope="function")
def get_default_test_beam_material() -> Callable:
    """Return a function to create a default beam material for testing
    purposes.

    Returns:
        A function that creates a default beam material.
    """

    def _get_default_test_beam_material(material_type: str = "base", **kwargs):
        """Return a default material for testing purposes.

        Args:
            material_type: The type of beam material to return.

        Returns:
            A material object corresponding to the specified beam type.
        """

        if material_type == "base":
            return MaterialBeamBase(radius=1.0, **kwargs)

        elif material_type == "reissner":
            return MaterialReissner(
                radius=1.0, youngs_modulus=1.0, nu=0.3, density=1.0, **kwargs
            )

        else:
            raise ValueError(f"Unknown beam type: {material_type}")

    return _get_default_test_beam_material


@pytest.fixture(scope="function")
def get_default_test_solid_material() -> Callable:
    """Return a function to create a default solid material for testing
    purposes.

    Args:
        material_type: The type of solid material to return.

    Returns:
        A function that creates a default solid material.
    """

    def _get_default_test_solid_material(
        material_type: str = "st_venant_kirchhoff",
    ):
        """Return a default solid material for testing purposes.

        Args:
            material_type: The type of solid material to return.

        Returns:
            A material object corresponding to the specified solid material type.
        """

        if material_type == "st_venant_kirchhoff":
            return MaterialStVenantKirchhoff(youngs_modulus=1.0, nu=0.3, density=1.0)

        elif material_type == "2d_shell":
            return MaterialSolid(
                material_string="MAT_Kirchhoff_Love_shell",
                data={"YOUNG_MODULUS": 10.0, "POISSON_RATIO": 0.3, "THICKNESS": 0.05},
            )

        elif material_type == "solid_nested":
            return MaterialSolid(
                material_string="MAT_ElastHyper",
                data={
                    "NUMMAT": 2,
                    "MATIDS": [
                        MaterialSolid(
                            material_string="ELAST_CoupSVK",
                            data={"YOUNG": 1.0, "NUE": 0.0},
                        ),
                        MaterialSolid(
                            material_string="MAT_ElastHyper",
                            data={
                                "NUMMAT": 1,
                                "MATIDS": [
                                    MaterialSolid(
                                        material_string="ELAST_CoupSVK",
                                        data={"YOUNG": 2.0, "NUE": 0.0},
                                    )
                                ],
                                "DENS": 1.0,
                            },
                        ),
                    ],
                    "DENS": 1.0,
                },
            )

        else:
            raise ValueError(f"Unknown solid material type: {material_type}")

    return _get_default_test_solid_material


@pytest.fixture(scope="function")
def get_default_test_solid_element_description() -> Callable:
    """Return a function to create a default solid element description for
    testing purposes.

    Args:
        element_type: The type of solid element to return.

    Returns:
        A function that creates a default solid element description.
    """

    def _get_default_test_solid_element_description(
        element_type: str = "2d_solid",
    ):
        """Return a default solid element description for testing purposes.

        Args:
            element_type: The type of solid element to return.

        Returns:
            A function that creates a default solid element description.
        """

        if element_type == "2d_solid":
            return {
                "KINEM": "nonlinear",
                "EAS": "none",
                "THICK": 1.0,
                "STRESS_STRAIN": "plane_strain",
                "GP": [3, 3],
            }

        elif element_type == "2d_shell":
            return {"type": "SHELL_KIRCHHOFF_LOVE_NURBS", "GP": [3, 3]}

        elif element_type == "3d_solid":
            return {"KINEM": "nonlinear"}

        else:
            raise ValueError(f"Unknown solid element type: {element_type}")

    return _get_default_test_solid_element_description



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file has utility functions for handling NURBS."""

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.core.rotation import Rotation as _Rotation


def ensure_3d_splinepy_object(splinepy_obj) -> None:
    """Ensure that a given splinepy object has 3D control points.

    Args:
        splinepy_obj:
            Splinepy object to ensure 3D, if this is an object with fewer than
            3 dimensions, it will be converted to 3D by adding the missing
            coordinates with a value of 0.
    """

    control_points_dim = splinepy_obj.control_points.shape[1]
    if control_points_dim == 3:
        pass
    elif control_points_dim < 3:
        # Add "empty" coordinates to make control points 3D
        control_points_3d = _np.zeros([len(splinepy_obj.control_points), 3])
        control_points_3d[:, :control_points_dim] = splinepy_obj.control_points
        splinepy_obj.control_points = control_points_3d
    else:
        raise ValueError(
            f"Splinepy object must be 1D, 2D or 3D, but has {control_points_dim} dimensions."
        )


def translate_splinepy(splinepy_obj, vector) -> None:
    """Translate a splinepy object by a vector.

    Args:
        vector: _np.array, list
            2D/3D vector to translate the splinepy object.
    """

    if not len(vector) == splinepy_obj.control_points.shape[1]:
        raise ValueError(
            f"Dimensions of translation vector and splinepy object do not match: {len(vector)} != {splinepy_obj.control_points.shape[1]}"
        )

    for point in splinepy_obj.control_points:
        point += vector


def rotate_splinepy(splinepy_obj, rotation: _Rotation, origin=None) -> None:
    """Rotate a splinepy object by a rotation object."""

    rotation_matrix = rotation.get_rotation_matrix()

    dimension = splinepy_obj.control_points.shape[1]
    if dimension == 2:
        if not _np.allclose(
            rotation_matrix[2, :], [0, 0, 1], rtol=0, atol=_bme.eps_quaternion
        ) or not _np.allclose(
            rotation_matrix[:, 2], [0, 0, 1], rtol=0, atol=_bme.eps_quaternion
        ):
            raise ValueError(
                "Rotation vector must be in the x-y plane for 2D splinepy objects."
            )
        rotation_matrix = rotation_matrix[:2, :2]

    if origin is None:
        origin = _np.zeros(dimension)

    for i_point, point in enumerate(splinepy_obj.control_points):
        point_new = _np.dot(rotation_matrix, point - origin) + origin
        splinepy_obj.control_points[i_point] = point_new
